# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import json
import logging
from collections import defaultdict
from collections.abc import Mapping
from enum import Enum
from pathlib import Path
from typing import List, Optional
import yaml
from pydantic import BaseModel, ConfigDict
from gammapy.makers import MapDatasetMaker
from gammapy.utils.scripts import read_yaml, to_yaml, write_yaml
from gammapy.utils.types import AngleType, EnergyType, PathType, TimeType

__all__ = ["AnalysisConfig"]

CONFIG_PATH = Path(__file__).resolve().parent / "config"
DOCS_FILE = CONFIG_PATH / "docs.yaml"

log = logging.getLogger(__name__)


def deep_update(d, u):
    """Recursively update a nested dictionary.

    Taken from: https://stackoverflow.com/a/3233356/19802442
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class ReductionTypeEnum(str, Enum):
    spectrum = "1d"
    cube = "3d"


class FrameEnum(str, Enum):
    icrs = "icrs"
    galactic = "galactic"


class RequiredHDUEnum(str, Enum):
    events = "events"
    gti = "gti"
    aeff = "aeff"
    bkg = "bkg"
    edisp = "edisp"
    psf = "psf"
    rad_max = "rad_max"


class BackgroundMethodEnum(str, Enum):
    reflected = "reflected"
    fov = "fov_background"
    ring = "ring"


class SafeMaskMethodsEnum(str, Enum):
    aeff_default = "aeff-default"
    aeff_max = "aeff-max"
    edisp_bias = "edisp-bias"
    offset_max = "offset-max"
    bkg_peak = "bkg-peak"


class MapSelectionEnum(str, Enum):
    counts = "counts"
    exposure = "exposure"
    background = "background"
    psf = "psf"
    edisp = "edisp"


class GammapyBaseConfig(BaseModel):
    """Base configuration class.

    Provides Pydantic model settings.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="forbid",
        validate_default=True,
        use_enum_values=True,
    )

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"


class SkyCoordConfig(GammapyBaseConfig):
    """Configuration for `~astropy.coordinates.SkyCoord`.

    Attributes
    ----------
    frame : str
        Coordinate frame (e.g., 'icrs', 'galactic').
    lon : `astropy.units.Quantity`
        Longitude or right ascension.
    lat : `astropy.units.Quantity`
        Latitude or declination.
    """

    frame: Optional[FrameEnum] = None
    lon: Optional[AngleType] = None
    lat: Optional[AngleType] = None


class EnergyAxisConfig(GammapyBaseConfig):
    min: Optional[EnergyType] = None
    max: Optional[EnergyType] = None
    nbins: Optional[int] = None


class SpatialCircleConfig(GammapyBaseConfig):
    frame: Optional[FrameEnum] = None
    lon: Optional[AngleType] = None
    lat: Optional[AngleType] = None
    radius: Optional[AngleType] = None


class EnergyRangeConfig(GammapyBaseConfig):
    min: Optional[EnergyType] = None
    max: Optional[EnergyType] = None


class TimeRangeConfig(GammapyBaseConfig):
    start: Optional[TimeType] = None
    stop: Optional[TimeType] = None


class FluxPointsConfig(GammapyBaseConfig):
    """Configuration for the `~gammapy.estimators.FluxPointsEstimator`.

    Attributes
    ----------
    energy : dict
        Energy binning for the light curve. Should contain the following keys
        'min' and 'max' (with energy quantities) and 'nbins'.
    source : str
        Source name.
    parameters : dict
        Optional parameters such as selection filters, options are:
        "all", "errn-errp", "ul", "scan"
    """

    energy: EnergyAxisConfig = EnergyAxisConfig()
    source: str = "source"
    parameters: dict = {"selection_optional": "all"}


class LightCurveConfig(GammapyBaseConfig):
    """Configuration for the `~gammapy.estimators.LightCurveEstimator`.

    Attributes
    ----------
    time_intervals : dict
        Time intervals for the light curve. Should contain the following keys
        'start' and 'stop' as time quantities.
    energy_edges : dict
        Energy binning for the light curve. Should contain the following keys
        'min' and 'max' (with energy quantities) and 'nbins'.
    source : str
        Source name.
    parameters : dict
        Optional parameters such as selection filters.
    """

    time_intervals: TimeRangeConfig = TimeRangeConfig()
    energy_edges: EnergyAxisConfig = EnergyAxisConfig()
    source: str = "source"
    parameters: dict = {"selection_optional": "all"}


class FitConfig(GammapyBaseConfig):
    """Configuration for model fitting.

    Attributes
    ----------
    fit_range : dict
        Energy range used during the fit. Should contain the following keys
        'min' and 'max' (with energy quantities).
    """

    fit_range: EnergyRangeConfig = EnergyRangeConfig()


class ExcessMapConfig(GammapyBaseConfig):
    """Configuration for the `~gammapy.estimators.ExcessMapEstimator`.

    Attributes
    ----------
    correlation_radius : `astropy.units.Quantity`
        Radius for correlation/smoothing.
    parameters : dict
        Optional configuration parameters.
    energy_edges : dict
        Energy binning for map creation. Should contain the following keys
        'min' and 'max' (with energy quantities) and 'nbins'.
    """

    correlation_radius: AngleType = "0.1 deg"
    parameters: dict = {}
    energy_edges: EnergyAxisConfig = EnergyAxisConfig()


class BackgroundConfig(GammapyBaseConfig):
    """Configuration for background modeling.

    Attributes
    ----------
    method : list of str
        Background estimation method, the available options are:
        "reflected", "fov_background", "ring"
    exclusion : str
        Path to an exclusion region file.
    parameters : dict
        Additional options or method parameters.
    """

    method: Optional[BackgroundMethodEnum] = None
    exclusion: Optional[PathType] = None
    parameters: dict = {}


class SafeMaskConfig(GammapyBaseConfig):
    """Configuration for `~gammapy.makers.SafeMaskMaker`.

    Attributes
    ----------
    methods : list of str
        Masking methods to apply, the available options are:
        "aeff-default", "aeff-max", "edisp-bias", "offset-max", "bkg-peak"
    parameters : dict
        Additional configuration parameters.
    """

    methods: List[SafeMaskMethodsEnum] = [SafeMaskMethodsEnum.aeff_default]
    parameters: dict = {}


class EnergyAxesConfig(GammapyBaseConfig):
    energy: EnergyAxisConfig = EnergyAxisConfig(min="1 TeV", max="10 TeV", nbins=5)
    energy_true: EnergyAxisConfig = EnergyAxisConfig(
        min="0.5 TeV", max="20 TeV", nbins=16
    )


class SelectionConfig(GammapyBaseConfig):
    offset_max: AngleType = "2.5 deg"


class WidthConfig(GammapyBaseConfig):
    width: AngleType = "5 deg"
    height: AngleType = "5 deg"


class WcsConfig(GammapyBaseConfig):
    """Configuration for the WCS geometry.

    Attributes
    ----------
    skydir : `SkyCoordConfig`
        Sky coordinates configuration.
    binsize : `astropy.units.Quantity`
        Pixel size in degrees.
    width : dict
        Spatial extent of the map. Should contain one or both of the keys
        'width' and 'height', with angular quantities as values.
    binsize_irf : `astropy.units.Quantity`
        Pixel size for IRF-related maps.
    """

    skydir: SkyCoordConfig = SkyCoordConfig()
    binsize: AngleType = "0.02 deg"
    width: WidthConfig = WidthConfig()
    binsize_irf: AngleType = "0.2 deg"


class GeomConfig(GammapyBaseConfig):
    """Configuration for geometry.

    Attributes
    ----------
    wcs : `WcsConfig`
        WCS geometry configuration.
    selection : `astropy.units.Quantity`
        The only option here is 'offset_max`.
    axes : `EnergyAxesConfig`
        Configuration of energy axes.
    """

    wcs: WcsConfig = WcsConfig()
    selection: SelectionConfig = SelectionConfig()
    axes: EnergyAxesConfig = EnergyAxesConfig()


class DatasetsConfig(GammapyBaseConfig):
    """Configuration for dataset reduction.

    Attributes
    ----------
    type : str
        Type of dataset to create (e.g., 'spectrum').
    stack : bool
        Whether to stack observations.
    geom : `GeomConfig`
        Geometry configuration.
    map_selection : list of str
        Select which maps to make, the available options are:
        'counts', 'exposure', 'background', 'psf', 'edisp'.
    background : `BackgroundConfig`
        Background configuration.
    safe_mask : `SafeMaskConfig`
        Safe mask configuration.
    on_region : dict
        ON-region definition for spectral extraction. Should contain the following keys
        'frame', 'lat', 'lon' and 'radius', the latter three as angle quantities.
    containment_correction : bool
        Whether to apply containment correction.
    """

    type: ReductionTypeEnum = ReductionTypeEnum.spectrum
    stack: bool = True
    geom: GeomConfig = GeomConfig()
    map_selection: List[MapSelectionEnum] = MapDatasetMaker.available_selection
    background: BackgroundConfig = BackgroundConfig()
    safe_mask: SafeMaskConfig = SafeMaskConfig()
    on_region: SpatialCircleConfig = SpatialCircleConfig()
    containment_correction: bool = True


class ObservationsConfig(GammapyBaseConfig):
    """Configuration for `~gammapy.data.Observations`.

    Attributes
    ----------
    datastore : str
        Path to the data store.
    obs_ids : list of int
        List of observation IDs.
    obs_file : str
        Path to a YAML observation file.
    obs_cone : dict
        Cone selection for observations. Should contain the following keys
        'frame', 'lat', 'lon' and 'radius', the latter three as angle quantities.
    obs_time : dict
        Observation time filtering.  Should contain the following keys
        'start' and 'stop' as time quantities.
    required_irf : list
        Required IRF components, options are "aeff", "edisp", "psf", "bkg"
    """

    datastore: PathType = Path("$GAMMAPY_DATA/hess-dl3-dr1/")
    obs_ids: List[int] = []
    obs_file: Optional[PathType] = None
    obs_cone: SpatialCircleConfig = SpatialCircleConfig()
    obs_time: TimeRangeConfig = TimeRangeConfig()
    required_irf: List[RequiredHDUEnum] = ["aeff", "edisp", "psf", "bkg"]


class LogConfig(GammapyBaseConfig):
    """Configuration for logging.

    Attributes
    ----------
    level : str
        Logging level (e.g., 'info', 'debug').
    filename : str
        Log file path.
    filemode : str
        File mode ('w' for overwrite, 'a' for append).
    format : str
        Logging format string.
    datefmt : str
        Format for timestamps.
    """

    level: str = "info"
    filename: Optional[PathType] = None
    filemode: Optional[str] = None
    format: Optional[str] = None
    datefmt: Optional[str] = None


class GeneralConfig(GammapyBaseConfig):
    """Top-level general configuration.

    Attributes
    ----------
    log : `LogConfig`
        Logging configuration.
    outdir : str
        Output directory.
    n_jobs : int
        Number of parallel jobs.
    datasets_file : str
        Path to datasets config file.
    models_file : str
        Path to models config file.
    """

    log: LogConfig = LogConfig()
    outdir: str = "."
    n_jobs: int = 1
    datasets_file: Optional[PathType] = None
    models_file: Optional[PathType] = None


class AnalysisConfig(GammapyBaseConfig):
    """Gammapy analysis configuration.

    This class defines the full analysis configuration schema, organised into different sections.
    It can be read from or written to a YAML file using `.read()` and `.write()`, respectively.

    Attributes
    ----------
    general : `GeneralConfig`
        General settings for output, logging, and file paths.
    observations : `ObservationsConfig`
        Settings for the `~gammapy.data.Observation` selection including IDs, regions, and time filters.
    datasets : `DatasetsConfig`
        Settings for the `~gammapy.datasets.Dataset` Including but not limited to geometry (`GeomConfig`),
        background (`BackgroundConfig`), safe mask (`SafeMaskConfig`), and stacking.
    fit : `FitConfig`
        Configuration for the `~gammapy.modeling.Fit` strategy and global fit energy range.
    flux_points : `FluxPointsConfig`
        Configuration for the `~gammapy.estimators.FluxPointsEstimator`.
    excess_map : `ExcessMapConfig`
        Configuration for the `~gammapy.estimators.ExcessMapEstimator`.
    light_curve : `LightCurveConfig`
        Configuration for the `~gammapy.estimators.LightCurveEstimator`.

    Examples
    --------
    Read from a yaml file::

    >>> from gammapy.analysis import AnalysisConfig
    >>> config = AnalysisConfig.read("config.yaml") # doctest: +SKIP
    >>> print(config.datasets.geom) # doctest: +SKIP

    Create from scratch

    >>> config = AnalysisConfig()
    >>> config.observations.datastore = "$GAMMAPY_DATA/hess-dl3-dr1"
    >>> config.observations.obs_cone = {"frame": "icrs", "lon": "83.633 deg", "lat": "22.014 deg", "radius": "5 deg"}
    >>> print(config.observations.obs_cone.lat.deg)
    22.014
    """

    general: GeneralConfig = GeneralConfig()
    observations: ObservationsConfig = ObservationsConfig()
    datasets: DatasetsConfig = DatasetsConfig()
    fit: FitConfig = FitConfig()
    flux_points: FluxPointsConfig = FluxPointsConfig()
    excess_map: ExcessMapConfig = ExcessMapConfig()
    light_curve: LightCurveConfig = LightCurveConfig()

    def __str__(self):
        """Display settings in pretty YAML format."""
        info = self.__class__.__name__ + "\n\n\t"
        data = self.to_yaml()
        data = data.replace("\n", "\n\t")
        info += data
        return info.expandtabs(tabsize=4)

    @classmethod
    def read(cls, path):
        """Read from YAML file.

        Parameters
        ----------
        path : str
            input filepath
        """
        config = read_yaml(path)
        config.pop("metadata", None)
        return AnalysisConfig(**config)

    @classmethod
    def from_yaml(cls, config_str):
        """Create from YAML string.

        Parameters
        ----------
        config_str : str
            yaml str
        """
        settings = yaml.safe_load(config_str)
        return AnalysisConfig(**settings)

    def write(self, path, overwrite=False):
        """Write to YAML file.

        Parameters
        ----------
        path : `pathlib.Path` or str
            Path to write files.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        """
        yaml_str = self.to_yaml()
        write_yaml(yaml_str, path, overwrite=overwrite)

    def to_yaml(self):
        """Convert to YAML string."""
        data = json.loads(self.model_dump_json())
        return to_yaml(data)

    def set_logging(self):
        """Set logging config.

        Calls ``logging.basicConfig``, i.e. adjusts global logging state.
        """
        self.general.log.level = self.general.log.level.upper()
        logging.basicConfig(**self.general.log.model_dump())
        log.info("Setting logging config: {!r}".format(self.general.log.model_dump()))

    def update(self, config=None):
        """Update config with provided settings.

        Parameters
        ----------
        config : str or `AnalysisConfig` object, optional
            Configuration settings provided in dict() syntax. Default is None.
        """
        if isinstance(config, str):
            other = AnalysisConfig.from_yaml(config)
        elif isinstance(config, AnalysisConfig):
            other = config
        else:
            raise TypeError(f"Invalid type: {config}")

        config_new = deep_update(
            self.model_dump(exclude_defaults=True),
            other.model_dump(exclude_defaults=True),
        )
        return AnalysisConfig(**config_new)

    @staticmethod
    def _get_doc_sections():
        """Return dictionary with commented docs from docs file."""
        doc = defaultdict(str)
        with open(DOCS_FILE) as f:
            for line in filter(lambda line: not line.startswith("---"), f):
                line = line.strip("\n")
                if line.startswith("# Section: "):
                    keyword = line.replace("# Section: ", "")
                doc[keyword] += line + "\n"
        return doc



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy import units as u
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.table import Table
from gammapy.data import Observation
from gammapy.data.pointing import FixedPointingInfo
from gammapy.datasets import (
    Datasets,
    FluxPointsDataset,
    MapDataset,
    SpectrumDatasetOnOff,
)
from gammapy.datasets.spectrum import SpectrumDataset
from gammapy.estimators import FluxPoints, FluxPointsEstimator
from gammapy.estimators.utils import get_rebinned_axis
from gammapy.irf import EDispKernelMap, EffectiveAreaTable2D, load_irf_dict_from_file
from gammapy.makers import MapDatasetMaker
from gammapy.makers.utils import make_map_exposure_true_energy
from gammapy.maps import MapAxis, RegionGeom, RegionNDMap, WcsGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    ExpCutoffPowerLawSpectralModel,
    FoVBackgroundModel,
    GaussianSpatialModel,
    Models,
    PiecewiseNormSpectralModel,
    PowerLawSpectralModel,
    SkyModel,
    TemplateNPredModel,
    TemplateSpatialModel,
)
from gammapy.utils import parallel
from gammapy.utils.testing import requires_data, requires_dependency


@pytest.fixture()
def fermi_datasets():
    filename = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_datasets.yaml"
    filename_models = "$GAMMAPY_DATA/fermi-3fhl-crab/Fermi-LAT-3FHL_models.yaml"
    return Datasets.read(filename=filename, filename_models=filename_models)


# TODO: use pre-generated data instead
def simulate_spectrum_dataset(model, random_state=0):
    energy_edges = np.logspace(-0.5, 1.5, 21) * u.TeV
    energy_axis = MapAxis.from_edges(energy_edges, interp="log", name="energy")
    energy_axis_true = energy_axis.copy(name="energy_true")

    aeff = EffectiveAreaTable2D.from_parametrization(energy_axis_true=energy_axis_true)

    bkg_model = SkyModel(
        spectral_model=PowerLawSpectralModel(
            index=2.5, amplitude="1e-12 cm-2 s-1 TeV-1"
        ),
        name="background",
    )
    bkg_model.spectral_model.amplitude.frozen = True
    bkg_model.spectral_model.index.frozen = True

    geom = RegionGeom.create(region="icrs;circle(0, 0, 0.1)", axes=[energy_axis])
    acceptance = RegionNDMap.from_geom(geom=geom, data=1)
    edisp = EDispKernelMap.from_diagonal_response(
        energy_axis=energy_axis,
        energy_axis_true=energy_axis_true,
        geom=geom,
    )

    geom_true = RegionGeom.create(
        region="icrs;circle(0, 0, 0.1)", axes=[energy_axis_true]
    )
    exposure = make_map_exposure_true_energy(
        pointing=SkyCoord("0d", "0d"), aeff=aeff, livetime=100 * u.h, geom=geom_true
    )

    mask_safe = RegionNDMap.from_geom(geom=geom, dtype=bool)
    mask_safe.data += True

    acceptance_off = RegionNDMap.from_geom(geom=geom, data=5)
    dataset = SpectrumDatasetOnOff(
        name="test_onoff",
        exposure=exposure,
        acceptance=acceptance,
        acceptance_off=acceptance_off,
        edisp=edisp,
        mask_safe=mask_safe,
    )
    dataset.models = bkg_model
    bkg_npred = dataset.npred_signal()

    dataset.models = model
    dataset.fake(
        random_state=random_state,
        npred_background=bkg_npred,
    )
    return dataset


def create_fpe(model):
    model = SkyModel(spectral_model=model, name="source")
    dataset = simulate_spectrum_dataset(model)
    energy_edges = [0.1, 1, 10, 100] * u.TeV
    dataset.models = model
    fpe = FluxPointsEstimator(
        energy_edges=energy_edges,
        source="source",
        selection_optional="all",
        fit=Fit(backend="minuit", optimize_opts=dict(tol=0.2, strategy=1)),
    )
    fpe.norm.scan_n_values = 11
    datasets = [dataset]
    return datasets, fpe


def simulate_map_dataset(random_state=0, name=None):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )

    skydir = SkyCoord("0 deg", "0 deg", frame="galactic")
    pointing = FixedPointingInfo(fixed_icrs=skydir.icrs)
    energy_edges = np.logspace(-1, 2, 15) * u.TeV
    energy_axis = MapAxis.from_edges(edges=energy_edges, name="energy", interp="log")

    geom = WcsGeom.create(
        skydir=skydir, width=(4, 4), binsz=0.1, axes=[energy_axis], frame="galactic"
    )

    gauss = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.4 deg", frame="galactic"
    )
    pwl = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")
    skymodel = SkyModel(spatial_model=gauss, spectral_model=pwl, name="source")

    obs = Observation.create(
        pointing=pointing,
        livetime=1 * u.h,
        irfs=irfs,
        location=EarthLocation(lon="-70d18m58.84s", lat="-24d41m0.34s", height="2000m"),
    )
    empty = MapDataset.create(geom, name=name)
    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    dataset = maker.run(empty, obs)

    bkg_model = FoVBackgroundModel(dataset_name=dataset.name)

    dataset.models = [bkg_model, skymodel]
    dataset.fake(random_state=random_state)
    return dataset


@pytest.fixture(scope="session")
def fpe_map_pwl():
    dataset_1 = simulate_map_dataset(name="test-map-pwl")
    dataset_2 = dataset_1.copy(name="test-map-pwl-2")
    dataset_2.models = dataset_1.models

    dataset_2.mask_safe = RegionNDMap.from_geom(dataset_2.counts.geom, dtype=bool)

    energy_edges = [0.1, 1, 10, 100] * u.TeV
    datasets = [dataset_1, dataset_2]
    fpe = FluxPointsEstimator(
        energy_edges=energy_edges,
        source="source",
        selection_optional="all",
    )
    fpe.norm.scan_n_values = 3

    return datasets, fpe


@pytest.fixture(scope="session")
def fpe_map_pwl_ray():
    """duplicate of fpe_map_pwl to avoid fails due to execution order"""
    dataset_1 = simulate_map_dataset(name="test-map-pwl")
    dataset_2 = dataset_1.copy(name="test-map-pwl-2")
    dataset_2.models = dataset_1.models

    dataset_2.mask_safe = RegionNDMap.from_geom(dataset_2.counts.geom, dtype=bool)

    energy_edges = [0.1, 1, 10, 100] * u.TeV
    datasets = [dataset_1, dataset_2]
    fpe = FluxPointsEstimator(
        energy_edges=energy_edges,
        source="source",
        selection_optional="all",
    )
    fpe.norm.scan_n_values = 3

    return datasets, fpe


@pytest.fixture(scope="session")
def fpe_map_pwl_reoptimize():
    dataset = simulate_map_dataset()
    energy_edges = [1, 10] * u.TeV
    dataset.models.parameters["lon_0"].frozen = True
    dataset.models.parameters["lat_0"].frozen = True
    dataset.models.parameters["sigma"].frozen = True
    datasets = [dataset]
    fpe = FluxPointsEstimator(
        energy_edges=energy_edges,
        reoptimize=True,
        source="source",
    )
    fpe.norm.scan_values = [0.8, 1, 1.2]
    return datasets, fpe


@pytest.fixture(scope="session")
def fpe_pwl():
    return create_fpe(PowerLawSpectralModel())


@pytest.fixture(scope="session")
def fpe_ecpl():
    return create_fpe(ExpCutoffPowerLawSpectralModel(lambda_="1 TeV-1"))


def test_str(fpe_pwl):
    datasets, fpe = fpe_pwl
    assert "FluxPointsEstimator" in str(fpe)


def test_run_pwl(fpe_pwl, tmpdir):
    datasets, fpe = fpe_pwl

    fp = fpe.run(datasets)
    table = fp.to_table()

    actual = table["e_min"].data
    assert_allclose(actual, [0.316228, 1.0, 10.0], rtol=1e-5)

    actual = table["e_max"].data
    assert_allclose(actual, [1.0, 10.0, 31.622777], rtol=1e-5)

    actual = table["e_ref"].data
    assert_allclose(actual, [0.562341, 3.162278, 17.782794], rtol=1e-3)

    actual = table["ref_flux"].quantity
    desired = [2.162278e-12, 9.000000e-13, 6.837722e-14] * u.Unit("1 / (cm2 s)")
    assert_allclose(actual, desired, rtol=1e-3)

    actual = table["ref_dnde"].quantity
    desired = [3.162278e-12, 1.000000e-13, 3.162278e-15] * u.Unit("1 / (cm2 s TeV)")
    assert_allclose(actual, desired, rtol=1e-3)

    actual = table["ref_eflux"].quantity
    desired = [1.151293e-12, 2.302585e-12, 1.151293e-12] * u.Unit("TeV / (cm2 s)")
    assert_allclose(actual, desired, rtol=1e-3)

    actual = table["norm"].data
    assert_allclose(actual, [1.081434, 0.91077, 0.922176], rtol=1e-3)

    actual = table["norm_err"].data
    assert_allclose(actual, [0.066374, 0.061025, 0.179729], rtol=1e-2)

    actual = table["norm_errn"].data
    assert_allclose(actual, [0.065803, 0.060403, 0.171376], rtol=1e-2)

    actual = table["norm_errp"].data
    assert_allclose(actual, [0.06695, 0.061652, 0.18839], rtol=1e-2)

    actual = table["counts"].data.squeeze()
    assert_allclose(actual, [1490, 748, 43])

    actual = table["norm_ul"].data
    assert_allclose(actual, [1.216227, 1.035472, 1.316878], rtol=1e-2)

    actual = table["norm_sensitivity"].data
    assert_allclose(actual, [0.15901235, 0.13117017, 0.55880332], rtol=1e-2)

    actual = table["sqrt_ts"].data
    assert_allclose(actual, [18.568429, 18.054651, 7.057121], rtol=1e-2)

    actual = table["ts"].data
    assert_allclose(actual, [344.7866, 325.9704, 49.8029], rtol=1e-2)

    actual = table["stat"].data
    assert_allclose(actual, [2.76495, 13.11912, 3.70128], rtol=1e-2)

    actual = table["stat_null"].data
    assert_allclose(actual, [347.55159, 339.08952, 53.50424], rtol=1e-2)

    actual = table["norm_scan"][0][[0, 5, -1]]
    assert_allclose(actual, [0.2, 1.0, 5.0])

    actual = table["stat_scan"][0][[0, 5, -1]]
    assert_allclose(actual, [220.369, 4.301, 1881.626], rtol=1e-2)

    actual = table["npred"].data
    assert_allclose(actual, [[1492.966], [749.459], [43.105]], rtol=1e-3)

    actual = table["npred_excess"].data
    assert_allclose(actual, [[660.5625], [421.5402], [34.3258]], rtol=1e-3)

    actual = table.meta["UL_CONF"]
    assert_allclose(actual, 0.9544997)

    npred_excess_err = fp.npred_excess_err.data.squeeze()
    assert_allclose(npred_excess_err, [40.541334, 28.244024, 6.690005], rtol=1e-3)

    npred_excess_errp = fp.npred_excess_errp.data.squeeze()
    assert_allclose(npred_excess_errp, [40.838806, 28.549508, 7.013377], rtol=1e-3)

    npred_excess_errn = fp.npred_excess_errn.data.squeeze()
    assert_allclose(npred_excess_errn, [40.247313, 27.932033, 6.378465], rtol=1e-3)

    npred_excess_ul = fp.npred_excess_ul.data.squeeze()
    assert_allclose(npred_excess_ul, [742.87486, 479.169719, 49.019125], rtol=1e-3)

    # test GADF I/O
    fp.write(tmpdir / "test.fits")
    fp_new = FluxPoints.read(tmpdir / "test.fits")
    assert fp_new.meta["sed_type_init"] == "likelihood"

    # test datasets stat
    fp_dataset = FluxPointsDataset(data=fp, models=fp.reference_model)
    fp_dataset.stat_type = "chi2"
    assert_allclose(fp_dataset.stat_sum(), 3.82374, rtol=1e-4)
    fp_dataset.stat_type = "profile"
    assert_allclose(fp_dataset.stat_sum(), 3.790053, rtol=1e-4)
    fp_dataset.stat_type = "distrib"
    assert_allclose(fp_dataset.stat_sum(), 3.783325, rtol=1e-4)


def test_run_ecpl(fpe_ecpl, tmpdir):
    datasets, fpe = fpe_ecpl

    fp = fpe.run(datasets)

    table = fp.to_table()

    actual = table["ref_flux"].quantity
    desired = [9.024362e-13, 1.781341e-13, 1.260298e-18] * u.Unit("1 / (cm2 s)")
    assert_allclose(actual, desired, rtol=1e-3)

    actual = table["ref_dnde"].quantity
    desired = [1.351382e-12, 7.527318e-15, 2.523659e-22] * u.Unit("1 / (cm2 s TeV)")
    assert_allclose(actual, desired, rtol=1e-3)

    actual = table["ref_eflux"].quantity
    desired = [4.770557e-13, 2.787695e-13, 1.371963e-17] * u.Unit("TeV / (cm2 s)")
    assert_allclose(actual, desired, rtol=1e-3)

    actual = table["norm"].data
    assert_allclose(actual, [1.001683, 1.061821, 1.237512e03], rtol=1e-3)

    actual = table["norm_err"].data
    assert_allclose(actual, [1.386091e-01, 2.394241e-01, 3.259756e03], rtol=1e-2)

    actual = table["norm_errn"].data
    assert_allclose(actual, [1.374962e-01, 2.361246e-01, 2.888978e03], rtol=1e-2)

    actual = table["norm_errp"].data
    assert_allclose(actual, [1.397358e-01, 2.428481e-01, 3.716550e03], rtol=1e-2)

    actual = table["norm_ul"].data
    assert_allclose(actual, [1.283433e00, 1.555117e00, 9.698645e03], rtol=1e-2)

    actual = table["sqrt_ts"].data
    assert_allclose(actual, [7.678454, 4.735691, 0.399243], rtol=1e-2)

    # test GADF I/O
    fp.write(tmpdir / "test.fits")
    fp_new = FluxPoints.read(tmpdir / "test.fits")
    assert fp_new.meta["sed_type_init"] == "likelihood"


@requires_data()
def test_run_map_pwl(fpe_map_pwl, tmpdir):
    datasets, fpe = fpe_map_pwl
    fp = fpe.run(datasets)

    table = fp.to_table()

    actual = table["e_min"].data
    assert_allclose(actual, [0.1, 1.178769, 8.48342], rtol=1e-5)

    actual = table["e_max"].data
    assert_allclose(actual, [1.178769, 8.483429, 100.0], rtol=1e-5)

    actual = table["e_ref"].data
    assert_allclose(actual, [0.343332, 3.162278, 29.126327], rtol=1e-5)

    actual = table["norm"].data
    assert_allclose(actual, [0.974726, 0.96342, 0.994251], rtol=1e-2)

    actual = table["norm_err"].data
    assert_allclose(actual, [0.067637, 0.052022, 0.087059], rtol=3e-2)

    actual = table["counts"].data
    assert_allclose(actual, [[44611, 0], [1923, 0], [282, 0]])

    actual = table["norm_ul"].data
    assert_allclose(actual, [1.111852, 1.07004, 1.17829], rtol=1e-2)

    actual = table["sqrt_ts"].data
    assert_allclose(actual, [16.681221, 28.408676, 21.91912], rtol=1e-2)

    actual = table["norm_scan"][0]
    assert_allclose(actual, [0.2, 1.0, 5])

    actual = table["stat_scan"][0] - table["stat"][0]
    assert_allclose(actual, [1.628398e02, 1.452456e-01, 2.008018e03], rtol=1e-2)

    # test GADF I/O
    fp.write(tmpdir / "test.fits")
    fp_new = FluxPoints.read(tmpdir / "test.fits")
    assert fp_new.meta["sed_type_init"] == "likelihood"


@requires_data()
def test_run_map_pwl_reoptimize(fpe_map_pwl_reoptimize):
    datasets, fpe = fpe_map_pwl_reoptimize
    fpe = fpe.copy()
    fpe.selection_optional = ["scan"]

    fp = fpe.run(datasets)
    table = fp.to_table()

    actual = table["norm"].data
    assert_allclose(actual, 0.962368, rtol=1e-2)

    actual = table["norm_err"].data
    assert_allclose(actual, 0.053878, rtol=1e-2)

    actual = table["sqrt_ts"].data
    assert_allclose(actual, 25.196585, rtol=1e-2)

    actual = table["norm_scan"][0]
    assert_allclose(actual, [0.8, 1, 1.2])

    actual = table["stat_scan"][0] - table["stat"][0]
    assert_allclose(actual, [9.788123, 0.486066, 17.603708], rtol=1e-2)


def test_run_no_edip(fpe_pwl, tmpdir):
    datasets, fpe = fpe_pwl

    datasets = datasets.copy()

    datasets[0].models["source"].apply_irf["edisp"] = False
    fp = fpe.run(datasets)
    table = fp.to_table()
    actual = table["norm"].data
    assert_allclose(actual, [1.081434, 0.91077, 0.922176], rtol=1e-3)

    datasets[0].edisp = None

    fp = fpe.run(datasets)
    table = fp.to_table()
    actual = table["norm"].data
    assert_allclose(actual, [1.081434, 0.91077, 0.922176], rtol=1e-3)

    datasets[0].models["source"].apply_irf["edisp"] = True
    fp = fpe.run(datasets)
    table = fp.to_table()
    actual = table["norm"].data
    assert_allclose(actual, [1.081434, 0.91077, 0.922176], rtol=1e-3)


@requires_dependency("iminuit")
@requires_data()
def test_run_template_npred(fpe_map_pwl, tmpdir):
    datasets, fpe = fpe_map_pwl
    dataset = datasets[0]
    models = Models(dataset.models)
    model = TemplateNPredModel(dataset.background, datasets_names=[dataset.name])
    models.append(model)
    dataset.models = models
    dataset.background.data = 0

    fp = fpe.run(dataset)

    table = fp.to_table()
    actual = table["norm"].data
    assert_allclose(actual, [0.974726, 0.96342, 0.994251], rtol=1e-2)

    fpe.sum_over_energy_groups = True
    fp = fpe.run(dataset)

    table = fp.to_table()
    actual = table["norm"].data
    assert_allclose(actual, [0.955512, 0.965328, 0.995237], rtol=1e-2)


@requires_data()
def test_reoptimize_no_free_parameters(fpe_pwl, caplog):
    datasets, fpe = fpe_pwl
    fpe.reoptimize = True
    with pytest.raises(ValueError, match="No free parameters for fitting"):
        fpe.run(datasets)
    fpe.reoptimize = False


@requires_data()
def test_flux_points_estimator_no_norm_scan(fpe_pwl, tmpdir):
    datasets, fpe = fpe_pwl
    fpe.selection_optional = None

    fp = fpe.run(datasets)

    assert_allclose(fpe.fit.optimize_opts["tol"], 0.2)

    assert fp.sed_type_init == "likelihood"
    assert "stat_scan" not in fp._data

    # test GADF I/O
    fp.write(tmpdir / "test.fits")
    fp_new = FluxPoints.read(tmpdir / "test.fits")
    assert fp_new.meta["sed_type_init"] == "likelihood"


def test_no_likelihood_contribution():
    dataset = simulate_spectrum_dataset(
        SkyModel(spectral_model=PowerLawSpectralModel(), name="source")
    )

    dataset_2 = dataset.slice_by_idx(slices={"energy": slice(0, 5)})

    dataset.mask_safe = RegionNDMap.from_geom(dataset.counts.geom, dtype=bool)

    fpe = FluxPointsEstimator(energy_edges=[1.0, 3.0, 10.0] * u.TeV, source="source")
    table = fpe.run([dataset, dataset_2]).to_table()

    assert np.isnan(table["norm"]).all()
    assert np.isnan(table["norm_err"]).all()
    assert_allclose(table["counts"], 0)


def test_mask_shape():
    axis = MapAxis.from_edges([1.0, 3.0, 10.0], unit="TeV", interp="log", name="energy")
    geom_1 = WcsGeom.create(binsz=1, width=3, axes=[axis])
    geom_2 = WcsGeom.create(binsz=1, width=5, axes=[axis])

    dataset_1 = MapDataset.create(geom_1)
    dataset_2 = MapDataset.create(geom_2)
    dataset_1.gti = None
    dataset_2.gti = None
    dataset_1.psf = None
    dataset_2.psf = None
    dataset_1.edisp = None
    dataset_2.edisp = None
    dataset_2.mask_safe = None

    model = SkyModel(
        spectral_model=PowerLawSpectralModel(),
        spatial_model=GaussianSpatialModel(),
        name="source",
    )

    dataset_1.models = model
    dataset_2.models = model

    fpe = FluxPointsEstimator(energy_edges=[1, 10] * u.TeV, source="source")

    fp = fpe.run([dataset_2, dataset_1])
    table = fp.to_table()

    assert_allclose(table["counts"], 0)
    assert_allclose(table["npred"], 0)


def test_run_pwl_parameter_range(fpe_pwl):
    pl = PowerLawSpectralModel(amplitude="1e-16 cm-2s-1TeV-1")

    datasets, fpe = create_fpe(pl)

    fp = fpe.run(datasets)

    table_no_bounds = fp.to_table()

    fpe.norm.min = 0
    fpe.norm.max = 1e4
    fp = fpe.run(datasets)
    table_with_bounds = fp.to_table()

    actual = table_with_bounds["norm"].data
    assert_allclose(actual, [0.0, 0.0, 0.0], atol=1e-2)

    actual = table_with_bounds["norm_errp"].data
    assert_allclose(actual, [212.593368, 298.383045, 449.951747], rtol=1e-2)

    actual = table_with_bounds["norm_ul"].data
    assert_allclose(actual, [640.067576, 722.571371, 1414.22209], rtol=1e-2)

    actual = table_with_bounds["sqrt_ts"].data
    assert_allclose(actual, [0.0, 0.0, 0.0], atol=1e-2)

    actual = table_no_bounds["norm"].data
    assert_allclose(actual, [-511.76675, -155.75408, -853.547117], rtol=1e-3)

    actual = table_no_bounds["norm_err"].data
    assert_allclose(actual, [504.601499, 416.69248, 851.223077], rtol=1e-2)

    actual = table_no_bounds["norm_ul"].data
    assert_allclose(actual, [514.957128, 707.888477, 1167.105962], rtol=1e-2)

    actual = table_no_bounds["sqrt_ts"].data
    assert_allclose(actual, [-1.006081, -0.364848, -0.927819], rtol=1e-2)


def test_flux_points_estimator_small_edges():
    pl = PowerLawSpectralModel(amplitude="1e-11 cm-2s-1TeV-1")

    datasets, fpe = create_fpe(pl)

    fpe.energy_edges = datasets[0].counts.geom.axes["energy"].upsample(3).edges[1:4]
    fpe.selection_optional = []

    fp = fpe.run(datasets)

    assert_allclose(fp.ts.data[0, 0, 0], 2156.96959291)
    assert np.isnan(fp.ts.data[1, 0, 0])
    assert np.isnan(fp.npred.data[1, 0, 0])


def test_flux_points_recompute_ul(fpe_pwl):
    datasets, fpe = fpe_pwl
    fpe.selection_optional = ["all"]
    fp = fpe.run(datasets)
    assert_allclose(
        fp.flux_ul.data,
        [[[2.629819e-12]], [[9.319243e-13]], [[9.004449e-14]]],
        rtol=1e-3,
    )
    fp1 = fp.recompute_ul(n_sigma_ul=4)
    assert_allclose(
        fp1.flux_ul.data,
        [[[2.92877891e-12]], [[1.04993236e-12]], [[1.22089744e-13]]],
        rtol=1e-3,
    )
    assert fp1.meta["n_sigma_ul"] == 4

    # check that it returns a sensible value
    fp2 = fp.recompute_ul(n_sigma_ul=2)
    assert_allclose(fp2.flux_ul.data, fp.flux_ul.data, rtol=1e-2)


def test_flux_points_parallel_multiprocessing(fpe_pwl):
    datasets, fpe = fpe_pwl
    fpe.selection_optional = ["all"]
    fpe.n_jobs = 2
    assert fpe.n_jobs == 2

    fp = fpe.run(datasets)
    assert_allclose(
        fp.flux_ul.data,
        [[[2.629819e-12]], [[9.319243e-13]], [[9.004449e-14]]],
        rtol=1e-3,
    )


def test_global_n_jobs_default_handling():
    fpe = FluxPointsEstimator(energy_edges=[1, 3, 10] * u.TeV)

    assert fpe.n_jobs == 1

    parallel.N_JOBS_DEFAULT = 2
    assert fpe.n_jobs == 2

    fpe.n_jobs = 5
    assert fpe.n_jobs == 5

    fpe.n_jobs = None
    assert fpe.n_jobs == 2
    assert fpe._n_jobs is None

    parallel.N_JOBS_DEFAULT = 1
    assert fpe.n_jobs == 1


@requires_dependency("ray")
def test_flux_points_parallel_ray(fpe_pwl):
    datasets, fpe = fpe_pwl
    fpe.selection_optional = ["all"]
    fpe.parallel_backend = "ray"
    fpe.n_jobs = 2
    fp = fpe.run(datasets)
    assert_allclose(
        fp.flux_ul.data,
        [[[2.629819e-12]], [[9.319243e-13]], [[9.004449e-14]]],
        rtol=1e-3,
    )


@requires_dependency("ray")
def test_flux_points_parallel_ray_actor_spectrum(fpe_pwl):
    from gammapy.datasets.actors import DatasetsActor

    datasets, fpe = fpe_pwl
    with pytest.raises(TypeError):
        DatasetsActor(datasets)


@requires_data()
@requires_dependency("ray")
def test_flux_points_parallel_ray_actor_map(fpe_map_pwl_ray):
    from gammapy.datasets.actors import DatasetsActor

    datasets, fpe = fpe_map_pwl_ray
    actors = DatasetsActor(datasets)

    fp = fpe.run(actors)

    table = fp.to_table()

    actual = table["e_min"].data
    assert_allclose(actual, [0.1, 1.178769, 8.48342], rtol=1e-5)

    actual = table["e_max"].data
    assert_allclose(actual, [1.178769, 8.483429, 100.0], rtol=1e-5)

    actual = table["e_ref"].data
    assert_allclose(actual, [0.343332, 3.162278, 29.126327], rtol=1e-5)

    actual = table["norm"].data
    assert_allclose(actual, [0.974726, 0.96342, 0.994251], rtol=1e-2)

    actual = table["norm_err"].data
    assert_allclose(actual, [0.067637, 0.052022, 0.087059], rtol=3e-2)

    actual = table["counts"].data
    assert_allclose(actual, [[44611, 0], [1923, 0], [282, 0]])

    actual = table["norm_ul"].data
    assert_allclose(actual, [1.111852, 1.07004, 1.17829], rtol=1e-2)

    actual = table["sqrt_ts"].data
    assert_allclose(actual, [16.681221, 28.408676, 21.91912], rtol=1e-2)

    actual = table["norm_scan"][0]
    assert_allclose(actual, [0.2, 1.0, 5])

    actual = table["stat_scan"][0] - table["stat"][0]
    assert_allclose(actual, [1.628398e02, 1.452456e-01, 2.008018e03], rtol=1e-2)


def test_fpe_non_aligned_energy_axes():
    energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=10)
    geom_1 = RegionGeom.create("icrs;circle(0, 0, 0.1)", axes=[energy_axis])
    dataset_1 = SpectrumDataset.create(geom=geom_1)

    energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=7)
    geom_2 = RegionGeom.create("icrs;circle(0, 0, 0.1)", axes=[energy_axis])
    dataset_2 = SpectrumDataset.create(geom=geom_2)

    fpe = FluxPointsEstimator(energy_edges=[1, 3, 10] * u.TeV)

    with pytest.raises(ValueError, match="must have aligned energy axes"):
        fpe.run(datasets=[dataset_1, dataset_2])


def test_fpe_non_uniform_datasets():
    energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=10)
    geom_1 = RegionGeom.create("icrs;circle(0, 0, 0.1)", axes=[energy_axis])
    dataset_1 = SpectrumDataset.create(
        geom=geom_1, meta_table=Table({"TELESCOP": ["CTA"]})
    )

    energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=10)
    geom_2 = RegionGeom.create("icrs;circle(0, 0, 0.1)", axes=[energy_axis])
    dataset_2 = SpectrumDataset.create(
        geom=geom_2, meta_table=Table({"TELESCOP": ["CTB"]})
    )

    fpe = FluxPointsEstimator(energy_edges=[1, 3, 10] * u.TeV)

    with pytest.raises(ValueError, match="same value of the 'TELESCOP' meta keyword"):
        fpe.run(datasets=[dataset_1, dataset_2])


@requires_data()
def test_flux_points_estimator_norm_spectral_model(fermi_datasets):
    energy_edges = [10, 30, 100, 300, 1000] * u.GeV

    model_ref = fermi_datasets.models["Crab Nebula"]
    estimator = FluxPointsEstimator(
        energy_edges=energy_edges,
        source="Crab Nebula",
        selection_optional=[],
        reoptimize=True,
    )
    flux_points = estimator.run(fermi_datasets[0])
    flux_points_dataset = FluxPointsDataset(data=flux_points, models=model_ref)
    flux_pred_ref = flux_points_dataset.flux_pred()

    models = Models([model_ref])
    geom = fermi_datasets[0].exposure.geom.to_image()
    energy_axis = MapAxis.from_energy_bounds(
        3 * u.GeV, 1.7 * u.TeV, nbin=30, per_decade=True, name="energy_true"
    )
    geom = geom.to_cube([energy_axis])

    model = models.to_template_sky_model(geom, name="test")
    fermi_datasets.models = [fermi_datasets[0].background_model, model]
    estimator = FluxPointsEstimator(
        energy_edges=energy_edges, source="test", selection_optional=[], reoptimize=True
    )
    flux_points = estimator.run(fermi_datasets[0])

    flux_points_dataset = FluxPointsDataset(data=flux_points, models=model)
    flux_pred = flux_points_dataset.flux_pred()
    assert_allclose(flux_pred, flux_pred_ref, rtol=2e-4)

    # test model 2d
    norms = (
        model.spatial_model.map.data.sum(axis=(1, 2))
        / model.spatial_model.map.data.sum()
    )
    model.spatial_model = TemplateSpatialModel(
        model.spatial_model.map.reduce_over_axes(), normalize=False
    )
    model.spectral_model = PiecewiseNormSpectralModel(geom.axes[0].center, norms)
    flux_points_dataset = FluxPointsDataset(data=flux_points, models=model)
    flux_pred = flux_points_dataset.flux_pred()
    assert_allclose(flux_pred, flux_pred_ref, rtol=2e-4)


@requires_data()
def test_fpe_diff_lengths():
    dataset = SpectrumDatasetOnOff.read(
        "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
    )
    dataset1 = SpectrumDatasetOnOff.read(
        "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23559.fits"
    )

    dataset.meta_table = Table(names=["NAME", "TELESCOP"], data=[["23523"], ["hess"]])
    dataset1.meta_table = Table(names=["NAME", "TELESCOP"], data=[["23559"], ["hess"]])
    dataset2 = Datasets([dataset, dataset1]).stack_reduce(name="dataset2")

    dataset3 = dataset1.copy()
    dataset3.meta_table = None

    pwl = PowerLawSpectralModel()

    datasets = Datasets([dataset1, dataset2, dataset3])

    datasets.models = SkyModel(spectral_model=pwl, name="test")
    energy_edges = [1, 2, 4, 10] * u.TeV
    fpe = FluxPointsEstimator(energy_edges=energy_edges, source="test")

    fp = fpe.run(datasets)

    assert_allclose(
        fp.dnde.data,
        [[[2.034323e-11]], [[3.39049716e-12]], [[2.96231326e-13]]],
        rtol=1e-3,
    )

    dataset4 = dataset1.copy()
    dataset4.meta_table = Table(names=["NAME", "TELESCOP"], data=[["23523"], ["not"]])
    datasets = Datasets([dataset1, dataset2, dataset3, dataset4])
    with pytest.raises(ValueError):
        fp = fpe.run(datasets)


def test_resample_axis():
    ds, fpe = create_fpe(PowerLawSpectralModel())
    flux_points = fpe.run(ds)
    axis_new = get_rebinned_axis(flux_points, method="fixed-bins", group_size=3)
    flux_new = flux_points.resample_axis(axis_new)
    assert_allclose(flux_new.flux.data, [[[3.1050191 * 1e-12]]], rtol=1e-5)


def test_warning_for_bad_model(caplog):
    ds, fpe = create_fpe(PowerLawSpectralModel())
    spectral_model = PowerLawSpectralModel()
    spectral_model.amplitude.value = -3e-14
    ds[0].models = SkyModel(spectral_model, name="source")
    fpe.run(ds)
    assert "WARNING" in [_.levelname for _ in caplog.records]
    assert (
        "Reference source model predicts negative flux. Results of estimator should be interpreted with caution"
        in [_.message for _ in caplog.records]
    )



import os
import logging
import click
from gammapy.scripts.info import cli_info
from gammapy.scripts.download import cli_download_datasets

log = logging.getLogger(__name__)

RELEASE = "latest"  # Or set to a specific release version if needed


@click.group(name="tutorial")
def cli_tutorial():
    """Gammapy tutorial helper commands."""


@cli_tutorial.command(name="setup")
@click.option(
    "--path",
    default="./gammapy-data",
    show_default=True,
    help="Path to download the datasets if needed.",
)
def cli_tutorial_setup(path):
    """Check tutorial setup and show environment info."""
    if "GAMMAPY_DATA" not in os.environ:
        log.info(f"GAMMAPY_DATA not set. Downloading to '{path}'...")
        cli_download_datasets.callback(out=path, release=RELEASE)
        os.environ["GAMMAPY_DATA"] = path
    else:
        log.info(f"GAMMAPY_DATA is already set to '{os.environ['GAMMAPY_DATA']}'")

    # Call the original info command
    cli_info.callback(system=True, version=True, dependencies=True, envvar=True)



"""Plot velocity distributions of Galactic sources."""

import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt
from gammapy.astro.population import velocity_distributions
from gammapy.maps.axes import UNIT_STRING_FORMAT
from gammapy.utils.random import normalize

velocity = np.linspace(10, 3000, 200) * u.km / u.s

ax = plt.subplot()

for key in velocity_distributions:
    model = velocity_distributions[key]()
    label = model.__class__.__name__
    x = velocity.value
    y = normalize(model, velocity[0].value, velocity[-1].value)(velocity.value)
    ax.plot(x, y, linestyle="-", label=label)

ax.set_xlim(velocity[0].value, velocity[-1].value)
ax.set_ylim(0, 0.005)
ax.set_xlabel(f"Velocity [{velocity.unit.to_string(UNIT_STRING_FORMAT)}]")
ax.set_ylabel(
    f"Probability Density [{((velocity.unit)**(-1)).to_string(UNIT_STRING_FORMAT)}]"
)
ax.semilogx()
plt.legend(prop={"size": 10})
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import astropy.io.fits as fits
import pytest

from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data
from ..io import EventListReader, EventListWriter


@requires_data()
def test_eventlist_reader_no_format():
    hess_events = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    fermi_events = "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-events.fits.gz"

    events = EventListReader().read(hess_events, format=None)
    assert len(events.table) == 11243

    events = EventListReader().read(fermi_events, format=None)
    assert len(events.table) == 32843


@requires_data()
def test_eventlist_reader_unkwnown_format(tmpdir):
    hess_events = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    )

    with fits.open(hess_events) as hdulist:
        hdu_events = hdulist["EVENTS"]
        hdu_events.header["HDUCLASS"] = "bad"

        hdulist.writeto(tmpdir / "tmp.fits")

    with pytest.raises(ValueError):
        EventListReader().read(tmpdir / "tmp.fits", format=None)


def test_eventlist_writer_unkwnown_format():
    with pytest.raises(ValueError):
        EventListWriter().to_hdu("tmp.fits", format="unknown")


@requires_data()
def test_eventlist_reader_empty_gadf_table():
    swgo_events = "$GAMMAPY_DATA/tests/format/swgo/map_irfs/DummyEvents.fits.gz"

    with pytest.raises(ValueError) as err:
        EventListReader().read(swgo_events, format="gadf")
    assert "ENERGY" in str(err.value)
    assert "TIME" in str(err.value)
    assert "RA" in str(err.value)
    assert "DEC" in str(err.value)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Dark matter spatial and spectral models."""
from gammapy.modeling.models import SPECTRAL_MODEL_REGISTRY
from .profiles import (
    BurkertProfile,
    DMProfile,
    EinastoProfile,
    IsothermalProfile,
    MooreProfile,
    NFWProfile,
    ZhaoProfile,
)
from .spectra import (
    DarkMatterAnnihilationSpectralModel,
    DarkMatterDecaySpectralModel,
    PrimaryFlux,
)
from .utils import JFactory

__all__ = [
    "DarkMatterAnnihilationSpectralModel",
    "DarkMatterDecaySpectralModel",
    "JFactory",
    "PrimaryFlux",
    "BurkertProfile",
    "DMProfile",
    "EinastoProfile",
    "IsothermalProfile",
    "MooreProfile",
    "NFWProfile",
    "ZhaoProfile",
]

SPECTRAL_MODEL_REGISTRY.append(DarkMatterAnnihilationSpectralModel)
SPECTRAL_MODEL_REGISTRY.append(DarkMatterDecaySpectralModel)



r"""
.. _broken-powerlaw-spectral-model:

Broken power law spectral model
===============================

This model parametrises a broken power law spectrum.

It is defined by the following equation:

.. math::
    \phi(E) = \phi_0 \cdot \begin{cases}
                          \left( \frac{E}{E_{break}} \right)^{-\Gamma1} & \text{if } E < E_{break} \\
                          \left( \frac{E}{E_{break}} \right)^{-\Gamma2} & \text{otherwise}
                         \end{cases}
    """

# %%
# Example plot
# ------------
# Here is an example plot of the model:

from astropy import units as u
import matplotlib.pyplot as plt
from gammapy.modeling.models import BrokenPowerLawSpectralModel, Models, SkyModel

energy_bounds = [0.1, 100] * u.TeV
model = BrokenPowerLawSpectralModel(
    index1=1.5,
    index2=2.5,
    amplitude="1e-12 TeV-1 cm-2 s-1",
    ebreak="1 TeV",
)
model.plot(energy_bounds)
plt.grid(which="both")

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(spectral_model=model, name="broken-power-law-model")
models = Models([model])

print(models.to_yaml())



"""
Flux Profile Estimation
=======================

Learn how to estimate flux profiles on a Fermi-LAT dataset.

Prerequisites
-------------

Knowledge of 3D data reduction and datasets used in Gammapy, see for
instance the first analysis tutorial.

Context
-------

A useful tool to study and compare the spatial distribution of flux in
images and data cubes is the measurement of flux profiles. Flux profiles
can show spatial correlations of gamma-ray data with e.g. gas maps or
other type of gamma-ray data. Most commonly flux profiles are measured
along some preferred coordinate axis, either radially distance from a
source of interest, along longitude and latitude coordinate axes or
along the path defined by two spatial coordinates.

Proposed Approach
-----------------

Flux profile estimation essentially works by estimating flux points for
a set of predefined spatially connected regions. For radial flux
profiles the shape of the regions are annuli with a common center, for
linear profiles it’s typically a rectangular shape.

We will work on a pre-computed `~gammapy.datasets.MapDataset` of Fermi-LAT data, use
`~regions.SkyRegion` to define the structure of the bins of the flux profile and
run the flux profile extraction using the `~gammapy.estimators.FluxProfileEstimator`

"""

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

# %matplotlib inline
import matplotlib.pyplot as plt

######################################################################
# Setup
# -----
#
from IPython.display import display
from gammapy.datasets import MapDataset
from gammapy.estimators import FluxPoints, FluxProfileEstimator
from gammapy.maps import RegionGeom
from gammapy.modeling.models import PowerLawSpectralModel
from gammapy.utils.regions import (
    make_concentric_annulus_sky_regions,
    make_orthogonal_rectangle_sky_regions,
)


######################################################################
# Read and Introduce Data
# -----------------------
#

dataset = MapDataset.read(
    "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc.fits.gz", name="fermi-dataset"
)


######################################################################
# This is what the counts image we will work with looks like:
#
counts_image = dataset.counts.sum_over_axes()
counts_image.smooth("0.1 deg").plot(stretch="sqrt")
plt.show()


######################################################################
# There are 400x200 pixels in the dataset and 11 energy bins between 10
# GeV and 2 TeV:
#

print(dataset.counts)


######################################################################
# Profile Estimation
# ------------------
#
# Configuration
# ~~~~~~~~~~~~~
#
# We start by defining a list of spatially connected regions along the
# galactic longitude axis. For this there is a helper function
# `~gammapy.utils.regions.make_orthogonal_rectangle_sky_regions`. The individual region bins
# for the profile have a height of 3 deg and in total there are 31 bins.
# Its starts from lon = 10 deg and goes to lon = 350 deg. In addition, we
# have to specify the ``wcs`` to take into account possible projections
# effects on the region definition:
#

regions = make_orthogonal_rectangle_sky_regions(
    start_pos=SkyCoord("9d", "0d", frame="galactic"),
    end_pos=SkyCoord("351d", "0d", frame="galactic"),
    wcs=counts_image.geom.wcs,
    height="3 deg",
    nbin=49,
)


######################################################################
# We can use the `~gammapy.maps.RegionGeom` object to illustrate the regions on top of
# the counts image:
#

geom = RegionGeom.create(region=regions)
ax = counts_image.smooth("0.1 deg").plot(stretch="sqrt")
geom.plot_region(ax=ax, color="w")
plt.show()


######################################################################
# Next we create the `~gammapy.estimators.FluxProfileEstimator`. For the estimation of the
# flux profile we assume a spectral model with a power-law shape and an
# index of 2.3
#

flux_profile_estimator = FluxProfileEstimator(
    regions=regions,
    spectral_model=PowerLawSpectralModel(index=2.3),
    energy_edges=[10, 2000] * u.GeV,
    selection_optional=["ul"],
)


######################################################################
# We can see the full configuration by printing the estimator object:
#

print(flux_profile_estimator)


######################################################################
# Run Estimation
# ~~~~~~~~~~~~~~
#
# Now we can run the profile estimation and explore the results:
#

profile = flux_profile_estimator.run(datasets=dataset)

print(profile)


######################################################################
# We can see the flux profile is represented by a `~gammapy.estimators.FluxPoints` object
# with a ``projected-distance`` axis, which defines the main axis the flux
# profile is measured along. The ``lon`` and ``lat`` axes can be ignored.
#
# Plotting Results
# ~~~~~~~~~~~~~~~~
#
# Let us directly plot the result using `~gammapy.estimators.FluxPoints.plot`:
#
ax = profile.plot(sed_type="dnde")
ax.set_yscale("linear")
plt.show()


######################################################################
# Based on the spectral model we specified above we can also plot in any
# other sed type, e.g. energy flux and define a different threshold when
# to plot upper limits:
#

profile.sqrt_ts_threshold_ul = 2

plt.figure()
ax = profile.plot(sed_type="eflux")
ax.set_yscale("linear")
plt.show()


######################################################################
# We can also plot any other quantity of interest, that is defined on the
# `~gammapy.estimators.FluxPoints` result object. E.g. the predicted total counts,
# background counts and excess counts:
#

quantities = ["npred", "npred_excess", "npred_background"]

fig, ax = plt.subplots()

for quantity in quantities:
    profile[quantity].plot(ax=ax, label=quantity.title())

ax.set_ylabel("Counts")
ax.legend()
plt.show()


######################################################################
# Serialisation and I/O
# ~~~~~~~~~~~~~~~~~~~~~
#
# The profile can be serialised using `~gammapy.estimators.FluxPoints.write`, given a
# specific format:
#

profile.write(
    filename="flux_profile_fermi.fits",
    format="profile",
    overwrite=True,
    sed_type="dnde",
)

profile_new = FluxPoints.read(filename="flux_profile_fermi.fits", format="profile")

ax = profile_new.plot()
ax.set_yscale("linear")
plt.show()


######################################################################
# The profile can be serialised to a `~astropy.table.Table` object
# using:
#

table = profile.to_table(format="profile", formatted=True)
display(table)


######################################################################
# No we can also estimate a radial profile starting from the Galactic
# center:
#

regions = make_concentric_annulus_sky_regions(
    center=SkyCoord("0d", "0d", frame="galactic"),
    radius_max="1.5 deg",
    nbin=11,
)


######################################################################
# Again we first illustrate the regions:
#
geom = RegionGeom.create(region=regions)
gc_image = counts_image.cutout(
    position=SkyCoord("0d", "0d", frame="galactic"), width=3 * u.deg
)
ax = gc_image.smooth("0.1 deg").plot(stretch="sqrt")
geom.plot_region(ax=ax, color="w")
plt.show()


######################################################################
# This time we define two energy bins and include the fit statistic
# profile in the computation:

flux_profile_estimator = FluxProfileEstimator(
    regions=regions,
    spectral_model=PowerLawSpectralModel(index=2.3),
    energy_edges=[10, 100, 2000] * u.GeV,
    selection_optional=["ul", "scan"],
)
######################################################################
# The configuration of the fit statistic profile is done throught the norm parameter:
flux_profile_estimator.norm.scan_values = np.linspace(-1, 5, 11)

######################################################################
# Now we can run the estimator,

profile = flux_profile_estimator.run(datasets=dataset)


######################################################################
# and plot the result:
#

profile.plot(axis_name="projected-distance", sed_type="flux")
plt.show()


######################################################################
# However because of the powerlaw spectrum the flux at high energies is
# much lower. To extract the profile at high energies only we can use:
#

profile_high = profile.slice_by_idx({"energy": slice(1, 2)})
plt.show()


######################################################################
# And now plot the points together with the likelihood profiles:
#

fig, ax = plt.subplots()
profile_high.plot(ax=ax, sed_type="eflux", color="tab:orange")
profile_high.plot_ts_profiles(ax=ax, sed_type="eflux")
ax.set_yscale("linear")
plt.show()

# sphinx_gallery_thumbnail_number = 2



"""Example plot showing stacking of two datasets."""

from astropy import units as u
from astropy.coordinates import SkyCoord
import matplotlib.pyplot as plt
from gammapy.data import Observation, observatory_locations
from gammapy.datasets import SpectrumDataset
from gammapy.datasets.map import MIGRA_AXIS_DEFAULT
from gammapy.irf import EffectiveAreaTable2D, EnergyDispersion2D
from gammapy.makers import SpectrumDatasetMaker
from gammapy.maps import MapAxis, RegionGeom
from gammapy.modeling.models import PowerLawSpectralModel, SkyModel

energy_true = MapAxis.from_energy_bounds(
    "0.1 TeV", "20 TeV", nbin=20, per_decade=True, name="energy_true"
)
energy_reco = MapAxis.from_energy_bounds("0.2 TeV", "10 TeV", nbin=10, per_decade=True)

aeff = EffectiveAreaTable2D.from_parametrization(
    energy_axis_true=energy_true, instrument="HESS"
)
offset_axis = MapAxis.from_bounds(0 * u.deg, 5 * u.deg, nbin=2, name="offset")

edisp = EnergyDispersion2D.from_gauss(
    energy_axis_true=energy_true,
    offset_axis=offset_axis,
    migra_axis=MIGRA_AXIS_DEFAULT,
    bias=0,
    sigma=0.2,
)

observation = Observation.create(
    obs_id=0,
    pointing=SkyCoord("0d", "0d", frame="icrs"),
    irfs={"aeff": aeff, "edisp": edisp},
    tstart=0 * u.h,
    tstop=0.5 * u.h,
    location=observatory_locations["hess"],
)

geom = RegionGeom.create("icrs;circle(0, 0, 0.1)", axes=[energy_reco])

stacked = SpectrumDataset.create(geom=geom, energy_axis_true=energy_true)

maker = SpectrumDatasetMaker(selection=["edisp", "exposure"])

dataset_1 = maker.run(stacked.copy(), observation=observation)
dataset_2 = maker.run(stacked.copy(), observation=observation)

pwl = PowerLawSpectralModel()
model = SkyModel(spectral_model=pwl, name="test-source")

dataset_1.mask_safe = geom.energy_mask(energy_min=2 * u.TeV)
dataset_2.mask_safe = geom.energy_mask(energy_min=0.6 * u.TeV)

dataset_1.models = model
dataset_2.models = model
dataset_1.counts = dataset_1.npred()
dataset_2.counts = dataset_2.npred()

stacked = dataset_1.copy(name="stacked")
stacked.stack(dataset_2)

stacked.models = model
npred_stacked = stacked.npred()

fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 4))

axes[0].set_title("Stacked Energy Dispersion Matrix")
axes[1].set_title("Predicted Counts")
stacked.edisp.get_edisp_kernel().plot_matrix(ax=axes[0])
npred_stacked.plot_hist(ax=axes[1], label="npred stacked")
stacked.counts.plot_hist(ax=axes[1], ls="--", label="stacked npred")
plt.legend()
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import os
from pathlib import Path
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.io import fits
from gammapy.data import DataStore
from gammapy.data.data_store import DataStoreMaker
from gammapy.irf import (
    Background3D,
    EffectiveAreaTable2D,
    EnergyDependentMultiGaussPSF,
    EnergyDispersion2D,
)
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import requires_data


@pytest.fixture()
def data_store():
    return DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")


@requires_data()
def test_data_store_hd_hap(data_store):
    """Test HESS HAP-HD data access."""
    obs = data_store.obs(obs_id=23523)

    assert obs.events.__class__.__name__ == "EventList"
    assert obs.gti.__class__.__name__ == "GTI"
    assert obs.aeff.__class__.__name__ == "EffectiveAreaTable2D"
    assert obs.edisp.__class__.__name__ == "EnergyDispersion2D"
    assert obs.psf.__class__.__name__ == "PSF3D"


@requires_data()
def test_data_store_from_dir():
    """Test the `from_dir` method."""
    data_store_rel_path = DataStore.from_dir(
        "$GAMMAPY_DATA/hess-dl3-dr1/", "hdu-index.fits.gz", "obs-index.fits.gz"
    )

    data_store_abs_path = DataStore.from_dir(
        "$GAMMAPY_DATA/hess-dl3-dr1/",
        "$GAMMAPY_DATA/hess-dl3-dr1/hdu-index.fits.gz",
        "$GAMMAPY_DATA/hess-dl3-dr1/obs-index.fits.gz",
    )

    assert "Data store" in data_store_rel_path.info(show=False)
    assert "Data store" in data_store_abs_path.info(show=False)


@requires_data()
def test_data_store_from_file(tmpdir):
    filename = "$GAMMAPY_DATA/hess-dl3-dr1/hdu-index.fits.gz"
    index_hdu = fits.open(make_path(filename))["HDU_INDEX"]

    filename = "$GAMMAPY_DATA/hess-dl3-dr1/obs-index.fits.gz"
    obs_hdu = fits.open(make_path(filename))["OBS_INDEX"]

    hdulist = fits.HDUList()
    hdulist.append(index_hdu)
    hdulist.append(obs_hdu)

    filename = tmpdir / "test-index.fits"
    hdulist.writeto(str(filename))

    data_store = DataStore.from_file(filename)

    assert data_store.obs_table["OBS_ID"][0] == 20136


@requires_data()
def test_data_store_get_observations(data_store, caplog):
    """Test loading data and IRF files via the DataStore"""
    observations = data_store.get_observations([23523, 23592])
    assert observations[0].obs_id == 23523
    observations = data_store.get_observations()
    assert len(observations) == 105

    with pytest.raises(ValueError):
        data_store.get_observations([11111, 23592])

    with caplog.at_level(logging.WARNING):
        observations = data_store.get_observations([11111, 23523], skip_missing=True)
        assert observations[0].obs_id == 23523
        assert "Skipping missing obs_id: 11111" in [_.message for _ in caplog.records]
    with caplog.at_level(logging.INFO):
        observations = data_store.get_observations([11111, 23523], skip_missing=True)
        assert "Observations selected: 1 out of 2." in [
            _.message for _ in caplog.records
        ]


@requires_data()
def test_broken_links_data_store(data_store):
    # Test that data_store without complete IRFs are properly loaded
    hdu_table = data_store.hdu_table
    index = np.where(hdu_table["OBS_ID"] == 23526)[0][0]
    hdu_table.remove_row(index)
    hdu_table._hdu_type_stripped = np.array([_.strip() for _ in hdu_table["HDU_TYPE"]])
    observations = data_store.get_observations(
        [23523, 23526], required_irf=["aeff", "edisp"]
    )
    assert len(observations) == 1

    with pytest.raises(ValueError):
        _ = data_store.get_observations([23523], required_irf=["xyz"])


@requires_data()
def test_data_store_copy_obs(tmp_path, data_store):
    data_store.copy_obs([23523, 23592], tmp_path, overwrite=True)

    substore = DataStore.from_dir(tmp_path)

    assert str(substore.hdu_table.base_dir) == str(tmp_path)
    assert len(substore.obs_table) == 2

    desired = data_store.obs(23523)
    actual = substore.obs(23523)

    assert str(actual.events.table) == str(desired.events.table)


@requires_data()
def test_data_store_copy_obs_subset(tmp_path, data_store):
    # Copy only certain HDU classes
    data_store.copy_obs([23523, 23592], tmp_path, hdu_class=["events"])

    substore = DataStore.from_dir(tmp_path)
    assert len(substore.hdu_table) == 2


@pytest.mark.xfail
@requires_data()
class TestDataStoreChecker:
    def setup_method(self):
        self.data_store = DataStore.from_dir("$GAMMAPY_DATA/cta-1dc/index/gps")

    def test_check_all(self):
        records = list(self.data_store.check())
        assert len(records) == 32


@pytest.fixture()
def data_store_dc1(monkeypatch):
    paths = [
        f"$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_{obs_id:06d}.fits"
        for obs_id in [110380, 111140, 111630, 111159]
    ]
    caldb_path = Path(os.environ["GAMMAPY_DATA"]) / Path("cta-1dc/caldb")
    monkeypatch.setenv("CALDB", str(caldb_path))
    return DataStore.from_events_files(paths)


@requires_data()
def test_data_store_from_events(data_store_dc1):
    # data_store_dc1 fixture is needed to set CALDB
    # Test that `DataStore.from_events_files` works.
    # The real tests for `DataStoreMaker` are below.
    path = "$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_110380.fits"
    data_store = DataStore.from_events_files([path])
    assert len(data_store.obs_table) == 1
    assert len(data_store.hdu_table) == 6
    assert data_store.obs_table.meta["MJDREFI"] == 51544

    path2 = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_025511.fits.gz"
    with pytest.raises(RuntimeError):
        _ = DataStore.from_events_files([path, path2])


@requires_data()
def test_read_events_cta_1dc(data_store_dc1):
    path = make_path("$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_110380.fits")
    info = DataStoreMaker.read_events_info(path)

    assert info["OBS_ID"] == 110380
    assert info["TSTART"] == 664502400.0 * u.s
    assert info["TSTOP"] == 664504192.0 * u.s
    assert info["ONTIME"] == 1800.0 * u.s
    assert info["LIVETIME"] == 1764.0 * u.s
    assert info["DEADC"] == 0.98000001907
    assert info["TELESCOP"] == "CTA"
    assert info["RA_PNT"] == 267.68121338 * u.deg
    assert info["DEC_PNT"] == -29.6075 * u.deg
    assert info["DATE-OBS"] == "2021-01-21"
    assert info["TIME-OBS"] == "11:58:51"
    assert info["DATE-END"] == "2021-01-21"
    assert info["TIME-END"] == "12:28:51"
    assert info["N_TELS"] == 0
    assert info["OBJECT"] == "Galactic Plane Survey"
    assert info["EVENT_COUNT"] == 106217
    assert info["CALDB"] == "1dc"
    assert info["IRF"] == "South_z20_50h"


@requires_data()
def test_read_events_info_hess():
    path = make_path(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    )
    info = DataStoreMaker.read_events_info(path)

    assert info["DATE-OBS"] == "2004-03-26"
    assert info["TIME-OBS"] == "02:57:46.184"
    assert info["DATE-END"] == "2004-03-26"


@requires_data()
def test_data_store_maker_obs_table(data_store_dc1):
    table = data_store_dc1.obs_table
    assert table.__class__.__name__ == "ObservationTable"
    assert len(table) == 4
    assert len(table.colnames) == 27
    assert table["CALDB"][0] == "1dc"
    assert table["IRF"][0] == "South_z20_50h"

    path = Path(table["IRF_FILENAME"][0])
    # checking str(path) would convert to windows path conventions on windows
    assert "$CALDB/data/cta/1dc/bcf/South_z20_50h/irf_file.fits" in repr(path)


@requires_data()
def test_data_store_maker_hdu_table(data_store_dc1):
    table = data_store_dc1.hdu_table
    assert table.__class__.__name__ == "HDUIndexTable"
    assert len(table) == 24
    hdu_class = ["events", "gti", "aeff_2d", "edisp_2d", "psf_3gauss", "bkg_3d"]
    assert list(data_store_dc1.hdu_table["HDU_CLASS"]) == 4 * hdu_class

    path = Path(table["FILE_DIR"][2])
    # checking str(path) would convert to windows path conventions on windows
    assert "$CALDB/data/cta/1dc/bcf/South_z20_50h" in repr(path)


@requires_data()
def test_data_store_maker_observation(data_store_dc1):
    """Check that one observation can be accessed OK"""

    obs = data_store_dc1.obs(110380)
    assert obs.obs_id == 110380

    assert obs.events.time[0].iso == "2021-01-21 12:00:03.045"
    assert obs.gti.time_start[0].iso == "2021-01-21 12:00:00.000"

    assert isinstance(obs.aeff, EffectiveAreaTable2D)
    assert isinstance(obs.bkg, Background3D)
    assert isinstance(obs.edisp, EnergyDispersion2D)
    assert isinstance(obs.psf, EnergyDependentMultiGaussPSF)


@requires_data("gammapy-data")
def test_data_store_fixed_rad_max():
    data_store = DataStore.from_dir("$GAMMAPY_DATA/joint-crab/dl3/magic")
    observations = data_store.get_observations(
        [5029748], required_irf=["aeff", "edisp"]
    )

    assert len(observations) == 1
    obs = observations[0]

    assert obs.rad_max is not None
    assert obs.rad_max.quantity.shape == (1, 1)
    assert u.allclose(obs.rad_max.quantity, np.sqrt(0.02) * u.deg)

    # test it also works with edisp (removing aeff)
    obs = data_store.get_observations([5029748], required_irf=["aeff", "edisp"])[0]
    obs.aeff = None
    assert obs.rad_max is not None
    assert obs.rad_max.quantity.shape == (1, 1)
    assert u.allclose(obs.rad_max.quantity, 0.1414213 * u.deg)

    # removing the last irf means we have no rad_max info
    obs = data_store.get_observations([5029748], required_irf=["aeff", "edisp"])[0]
    obs.aeff = None
    obs.edisp = None
    assert obs.rad_max is None


@requires_data()
def test_data_store_header_info_in_meta(data_store):
    """Test information from the obs index header is propagated into meta"""
    obs = data_store.obs(obs_id=23523)

    assert obs.meta.obs_info is not None
    # TODO: restructure test once all elements are in place


@requires_data()
def test_data_store_bad_name():
    with pytest.raises(IOError):
        DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/", "hdu-index.fits.gz", "bad")


@requires_data()
def test_data_store_doubled_id(caplog):
    """Test that doubled id in obs_id causes corresponding log.warning message"""

    # Using example provided in Issue #5943:
    datastore = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    datastore.get_observations(obs_id=[23523, 23523])
    assert "List of obs_id is not unique! Multiples are: [23523]" in [
        _.message for _ in caplog.records
    ]


@requires_data()
def test_data_store_from_dir_no_obs_index(caplog, tmpdir):
    """Test the `from_dir` method."""

    # Create small data_store and remove obs-index table
    DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/").copy_obs([23523, 23592], tmpdir)
    os.remove(tmpdir / "obs-index.fits.gz")

    data_store = DataStore.from_dir(tmpdir)

    obs = data_store.obs(23523)
    observations = data_store.get_observations()

    assert data_store.obs_table is None
    assert "No observation index table." in data_store.info(show=False)

    assert_allclose(obs.meta.location.lon.deg, 16.500222)
    assert len(observations) == 2

    test_dir = tmpdir / "test"
    os.mkdir(test_dir)
    data_store.copy_obs([23523], test_dir)
    data_store_copy = DataStore.from_dir(test_dir)
    assert len(data_store_copy.obs_ids) == 1
    assert data_store_copy.obs_table is None


@requires_data()
def test_data_store_required_irf_pointlike_fixed_rad_max():
    """Check behavior of the "point-like" option for data_store"""
    store = DataStore.from_dir("$GAMMAPY_DATA/joint-crab/dl3/magic")
    obs = store.get_observations([5029748], required_irf="point-like")
    assert len(obs) == 1
    assert u.allclose(obs[0].rad_max.quantity, np.sqrt(0.02) * u.deg)

    obs = store.get_observations([5029748], required_irf=["aeff", "edisp"])
    assert len(obs) == 1
    assert u.allclose(obs[0].rad_max.quantity, np.sqrt(0.02) * u.deg)


@requires_data()
def test_data_store_required_irf_pointlike_variable_rad_max():
    """Check behavior of the "point-like" option for data_store"""
    store = DataStore.from_dir("$GAMMAPY_DATA/magic/rad_max/data/")
    obs = store.get_observations(
        [5029747, 5029748], required_irf=["aeff", "edisp", "rad_max"]
    )
    assert len(obs) == 2
    assert obs[0].rad_max is not None

    obs = store.get_observations([5029747, 5029748], required_irf="point-like")
    assert len(obs) == 2
    assert obs[0].rad_max.quantity is not None


@requires_data()
def test_data_store_no_events():
    """Check behavior of the "point-like" option for data_store"""

    data_path = "$GAMMAPY_DATA/hawc/crab_events_pass4/"
    hdu_filename = "hdu-index-table-GP-no-events.fits.gz"
    obs_filename = "obs-index-table-GP-no-events.fits.gz"
    data_store = DataStore.from_dir(
        data_path, hdu_table_filename=hdu_filename, obs_table_filename=obs_filename
    )

    observations = data_store.get_observations(
        required_irf=["aeff", "psf", "edisp"], require_events=False
    )
    assert len(observations) == 3
    for obs in observations:
        assert not obs.events
        assert not obs.gti



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.modeling.models import (
    PRIOR_REGISTRY,
    GaussianPrior,
    GeneralizedGaussianPrior,
    Model,
    Models,
    SkyModel,
    UniformPrior,
    LogUniformPrior,
)
from gammapy.utils.testing import assert_quantity_allclose

TEST_PRIORS = [
    dict(
        name="gaussian",
        model=GaussianPrior(mu=4.0, sigma=1.0),
        prior_0=0.0 * u.Unit(""),
        prior_1=1.0 * u.Unit(""),
        val_at_0=17.837877,
        val_at_1=10.837877,
        inverse_cdf_at_0=-np.inf,
        inverse_cdf_at_1=np.inf,
    ),
    dict(
        name="uniform",
        model=UniformPrior(min=0.0, max=10),
        prior_0=0.0 * u.Unit(""),
        prior_1=11.0 * u.Unit(""),
        val_at_0=4.60517,
        val_at_1=np.inf,
        val_with_weight_2=9.21034,
        inverse_cdf_at_0=0.0,
        inverse_cdf_at_1=10.0,
    ),
    dict(
        name="loguniform",
        model=LogUniformPrior(min=1e-14, max=1e-10),
        prior_0=1e-10 * u.Unit(""),
        prior_1=1e-14 * u.Unit(""),
        val_at_0=-41.61104824714522,
        val_at_1=-60.03172899109759,
        inverse_cdf_at_0=1e-14,
        inverse_cdf_at_1=1e-10,
    ),
    dict(
        name="gennorm",
        model=GeneralizedGaussianPrior(mu=4, sigma=1.0, eta=0.5),
        prior_0=0.0 * u.Unit(""),
        prior_1=1.0 * u.Unit(""),
        val_at_0=17.837877,
        val_at_1=10.837877,
        inverse_cdf_at_0=-np.inf,
        inverse_cdf_at_1=np.inf,
    ),
]


@pytest.mark.parametrize("prior", TEST_PRIORS)
def test_prior_evaluation(prior):
    model = prior["model"]
    # Test the evaluation of the prior at specific points
    assert_allclose(model(prior["prior_0"]), prior["val_at_0"], rtol=1e-7)
    assert_allclose(model(prior["prior_1"]), prior["val_at_1"], rtol=1e-7)

    # Test the inverse_cdf at specific points
    value_0 = model._inverse_cdf(0)
    value_1 = model._inverse_cdf(1)
    assert_allclose(value_0, prior["inverse_cdf_at_0"], rtol=1e-7)
    assert_allclose(value_1, prior["inverse_cdf_at_1"], rtol=1e-7)


@pytest.mark.parametrize("prior", TEST_PRIORS)
def test_prior_parameters(prior):
    model = prior["model"]
    # Check that all parameters of the model have type 'prior'
    for p in model.parameters:
        assert p.type == "prior"


@pytest.mark.parametrize("prior", TEST_PRIORS)
def test_uniform_prior_weight(prior):
    prior = TEST_PRIORS[1]
    model = prior["model"]
    # Test the uniform prior with a specific weight
    model.weight = 2.0
    value_0_weight = model(0.0 * u.Unit(""))
    assert_allclose(value_0_weight, prior["val_with_weight_2"], rtol=1e-7)


def test_to_from_dict():
    prior = TEST_PRIORS[1]
    model = prior["model"]
    model.weight = 2.0
    model_dict = model.to_dict()
    # Here we reverse the order of parameters list to ensure assignment is correct
    model_dict["prior"]["parameters"].reverse()

    model_class = PRIOR_REGISTRY.get_cls(model_dict["prior"]["type"])
    new_model = model_class.from_dict(model_dict)

    assert isinstance(new_model, UniformPrior)

    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)
    assert_allclose(model.weight, new_model.weight, rtol=1e-7)

    new_model = Model.from_dict(model_dict)

    assert isinstance(new_model, UniformPrior)

    actual = [par.value for par in new_model.parameters]
    desired = [par.value for par in model.parameters]
    assert_quantity_allclose(actual, desired)
    assert_allclose(model.weight, new_model.weight, rtol=1e-7)


@pytest.mark.parametrize("prior", TEST_PRIORS)
def test_serialisation(prior, tmpdir):
    prior = TEST_PRIORS[1]
    model = SkyModel.create(spectral_model="pl", name="crab")
    model.spectral_model.amplitude.prior = prior["model"]
    models = Models([model])
    filename = str(tmpdir / "model_prior.yaml")
    models.write(filename)

    loaded_models = Models.read(filename)
    loaded_model = loaded_models[0]
    loaded_prior = loaded_model.spectral_model.amplitude.prior

    assert isinstance(loaded_prior, type(prior["model"]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from .asmooth import ASmoothMapEstimator
from .core import FluxMaps
from .excess import ExcessMapEstimator
from .ts import TSMapEstimator

__all__ = [
    "ASmoothMapEstimator",
    "ExcessMapEstimator",
    "FluxMaps",
    "TSMapEstimator",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from itertools import repeat
import numpy as np
import astropy.units as u
import gammapy.utils.parallel as parallel
from gammapy.data import GTI
from gammapy.datasets import Datasets
from gammapy.datasets.actors import DatasetsActor
from gammapy.maps import LabelMapAxis, Map, TimeMapAxis
from gammapy.modeling.models import Models
from gammapy.utils.pbar import progress_bar
from .core import FluxPoints
from .sed import FluxPointsEstimator

__all__ = ["LightCurveEstimator"]

log = logging.getLogger(__name__)


class LightCurveEstimator(FluxPointsEstimator):
    """Estimate light curve.

    The estimator will apply flux point estimation on the source model component to datasets
    in each of the provided time intervals.  The normalisation, ``norm``, is the only
    parameter of the source model left free to vary. Other model components
    can be left free to vary with the reoptimize option.

    If no time intervals are provided, the estimator will use the time intervals
    defined by the datasets GTIs.

    To be included in the estimation, the dataset must have their GTI fully
    overlapping a time interval.

    Time intervals without any dataset GTI fully overlapping will be dropped. They will not
    be stored in the final lightcurve `FluxPoints` object.

    Parameters
    ----------
    time_intervals : list of `~astropy.time.Time` objects
        Start and stop time for each interval to compute the LC.
    source : str or int, optional
        For which source in the model to compute the flux points.
        Default is 0, i.e. the first source of the models.
    atol : `~astropy.units.Quantity`, optional
        Tolerance value for time comparison with different scale. Default 1e-6 sec.
    n_sigma : float, optional
        Number of sigma to use for asymmetric error computation. Must be a positive value.
        Default is 1.
    n_sigma_ul : float, optional
        Number of sigma to use for upper limit computation. Must be a positive value.
        Default is 2.
    selection_optional : list of str, optional
        Which steps to execute. Available options are:

            * "all": all the optional steps are executed.
            * "errn-errp": estimate asymmetric errors.
            * "ul": estimate upper limits.
            * "scan": estimate fit statistic profiles.

        Default is None so the optional steps are not executed.
    energy_edges : list of `~astropy.units.Quantity`, optional
        Edges of the lightcurve energy bins. The resulting bin edges won't be exactly equal to the input ones,
        but rather the closest values to the energy axis edges of the parent dataset.
        Default is None: apply the estimator in each energy bin of the parent dataset.
        For further explanation see :ref:`estimators`.
    fit : `~gammapy.modeling.Fit`, optional
        Fit instance specifying the backend and fit options. If None, the `~gammapy.modeling.Fit` instance is created
        internally. Default is None.
    reoptimize : bool, optional
        If True the free parameters of the other models are fitted in each bin independently,
        together with the norm of the source of interest
        (but the other parameters of the source of interest are kept frozen).
        If False only the norm of the source of interest if fitted,
        and all other parameters are frozen at their current values.
        Default is False.
    stack_over_time_interval : bool, optional
        Whether to stack datasets within each time interval. Default is False.
        The predicted background counts from all datasets in the given interval
        will be stacked together, and the final background model (if any) will not
        have any free parameters. Available only if ``reoptimize`` is False.
    n_jobs : int, optional
        Number of processes used in parallel for the computation. Default is one,
        unless `~gammapy.utils.parallel.N_JOBS_DEFAULT` was modified. The number
        of jobs is limited to the number of physical CPUs.
    parallel_backend : {"multiprocessing", "ray"}, optional
        Which backend to use for multiprocessing. Defaults to `~gammapy.utils.parallel.BACKEND_DEFAULT`.
    norm : ~gammapy.modeling.Parameter` or dict, optional
        Norm parameter used for the fit.
        Default is None and a new parameter is created automatically,
        with value=1, name="norm", scan_min=0.2, scan_max=5, and scan_n_values = 11.
        By default, the min and max are not set and derived from the source model,
        unless the source model does not have one and only one norm parameter.
        If a dict is given the entries should be a subset of
        `~gammapy.modeling.Parameter` arguments.

    Examples
    --------
    For a usage example, see :doc:`/tutorials/analysis-time/light_curve` tutorial.

    Notes
    -----
    In case of failure of upper limits computation (e.g. nan), see the User Guide: :ref:`dropdown-UL`.
    """

    tag = "LightCurveEstimator"

    def __init__(
        self,
        time_intervals=None,
        atol="1e-6 s",
        stack_over_time_interval=False,
        **kwargs,
    ):
        self.time_intervals = time_intervals
        self.atol = u.Quantity(atol)
        self.stack_over_time_interval = stack_over_time_interval

        super().__init__(**kwargs)

    def run(self, datasets):
        """Run light curve extraction.

        Normalize integral and energy flux between emin and emax.

        Notes
        -----
        The progress bar can be displayed for this function.

        Parameters
        ----------
        datasets : list of `~gammapy.datasets.SpectrumDataset` or `~gammapy.datasets.MapDataset`
            Spectrum or Map datasets.

        Returns
        -------
        lightcurve : `~gammapy.estimators.FluxPoints`
            Light curve flux points.
        """
        if not isinstance(datasets, DatasetsActor):
            datasets = Datasets(datasets)

        if self.time_intervals is None:
            gti = datasets.gti
        else:
            gti = GTI.from_time_intervals(self.time_intervals)

        gti = gti.union(overlap_ok=False, merge_equal=False)

        rows = []
        valid_intervals = []
        parallel_datasets = []
        dataset_names = datasets.names
        for idx, (t_min, t_max) in enumerate(
            progress_bar(gti.time_intervals, desc="Time intervals selection")
        ):
            datasets_to_fit = datasets.select_time(
                time_min=t_min, time_max=t_max, atol=self.atol
            )

            if len(datasets_to_fit) == 0:
                log.info(
                    f"No Dataset for the time interval {t_min} to {t_max}. Skipping interval."
                )
                continue

            valid_intervals.append([t_min, t_max])
            if self.stack_over_time_interval and self.reoptimize:
                raise ValueError(
                    "It is not possible to stack the light curve if ``reoptimize=True``. You should set this value to False."
                )

            if self.stack_over_time_interval and not self.reoptimize:
                dataset_reduced = datasets_to_fit.stack_reduce(name="stacked")
                models = Models(datasets.models.copy())
                # Remove background models already applied in stack_reduce
                for model_name in models.background_models.values():
                    models.remove(model_name)
                for dataset in datasets:
                    models.reassign(dataset.name, dataset_reduced.name)
                dataset_reduced.models = models
                datasets_to_fit = Datasets([dataset_reduced])
                dataset_names = datasets_to_fit.names

            if self.n_jobs == 1:
                fp = self.estimate_time_bin_flux(datasets_to_fit, dataset_names)
                rows.append(fp)
            else:
                parallel_datasets.append(datasets_to_fit)

        if self.n_jobs > 1:
            self._update_child_jobs()
            rows = parallel.run_multiprocessing(
                self.estimate_time_bin_flux,
                zip(
                    parallel_datasets,
                    repeat(dataset_names),
                ),
                backend=self.parallel_backend,
                pool_kwargs=dict(processes=self.n_jobs),
                task_name="Time intervals",
            )

        if len(rows) == 0:
            raise ValueError("LightCurveEstimator: No datasets in time intervals")
        gti = GTI.from_time_intervals(valid_intervals)
        axis = TimeMapAxis.from_gti(gti=gti)
        return FluxPoints.from_stack(
            maps=rows,
            axis=axis,
        )

    @staticmethod
    def expand_map(m, dataset_names):
        """Expand map in dataset axis.

        Parameters
        ----------
        map : `~gammapy.maps.Map`
            Map to expand.
        dataset_names : list of str
            Dataset names.

        Returns
        -------
        map : `~gammapy.maps.Map`
            Expanded map.
        """
        label_axis = LabelMapAxis(labels=dataset_names, name="dataset")
        geom = m.geom.replace_axis(axis=label_axis)
        result = Map.from_geom(geom, data=np.nan)

        coords = m.geom.get_coord(sparse=True)
        result.set_by_coord(coords, vals=m.data)
        return result

    def estimate_time_bin_flux(self, datasets, dataset_names=None):
        """Estimate flux point for a single energy group.

        Parameters
        ----------
        datasets : `~gammapy.datasets.Datasets`
            List of dataset objects.
        dataset_names : list of str
            Dataset names.

        Returns
        -------
        result : `FluxPoints`
            Resulting flux points.
        """
        estimator = self.copy()
        estimator.n_jobs = self._n_child_jobs
        fp = estimator._run_flux_points(datasets)

        if dataset_names:
            for name in ["counts", "npred", "npred_excess"]:
                fp._data[name] = self.expand_map(
                    fp._data[name], dataset_names=dataset_names
                )
        return fp

    def _run_flux_points(self, datasets):
        return super().run(datasets)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import copy
import logging
from gammapy.utils.cache import cachemethod
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.io import fits
from astropy.table import QTable, Table
from astropy.utils import lazyproperty
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs.utils import (
    proj_plane_pixel_area,
    proj_plane_pixel_scales,
    wcs_to_celestial_frame,
)
from regions import (
    CompoundSkyRegion,
    PixCoord,
    PointSkyRegion,
    RectanglePixelRegion,
    Regions,
    SkyRegion,
)
import matplotlib.pyplot as plt
from gammapy.utils.regions import (
    compound_region_center,
    compound_region_to_regions,
    regions_to_compound_region,
)
from gammapy.visualization.utils import ARTIST_TO_LINE_PROPERTIES
from ..axes import MapAxes
from ..coord import MapCoord
from ..core import Map
from ..geom import Geom, pix_tuple_to_idx
from ..utils import _check_width
from ..wcs import WcsGeom

log = logging.getLogger(__name__)


__all__ = ["RegionGeom"]


def _parse_regions(regions):
    if isinstance(regions, str):
        regions = Regions.parse(data=regions, format="ds9")
    elif isinstance(regions, SkyRegion):
        regions = [regions]
    elif isinstance(regions, SkyCoord):
        regions = [PointSkyRegion(center=regions)]
    elif isinstance(regions, list) and len(regions) == 0:
        regions = None
    return regions


class RegionGeom(Geom):
    """Map geometry representing a region on the sky.

    The spatial component of the geometry is made up of a
    single pixel with an arbitrary shape and size. It can
    also have any number of non-spatial dimensions. This class
    represents the geometry for the `RegionNDMap` maps.

    Parameters
    ----------
    region : `~regions.SkyRegion`
        Region object.
    axes : list of `MapAxis`
        Non-spatial data axes.
    wcs : `~astropy.wcs.WCS`
        Optional WCS object to project the region if needed.
    binsz_wcs : `~astropy.units.Quantity`
        Angular bin size of the underlying `~WcsGeom` used to evaluate
        quantities in the region. Default is "0.1 deg". This default
        value is adequate for the majority of use cases. If a WCS object
        is provided, the input of ``binsz_wcs`` is overridden.

    Notes
    -----
    - For further explanation and examples, see :doc:`/user-guide/maps/regionmap`

    """

    is_regular = True
    is_allsky = False
    is_hpx = False
    is_region = True

    _slice_spatial_axes = slice(0, 2)
    _slice_non_spatial_axes = slice(2, None)
    projection = "TAN"

    def __init__(self, region, axes=None, wcs=None, binsz_wcs="0.1 deg"):
        self._region = region
        self._axes = MapAxes.from_default(axes, n_spatial_axes=2)
        self._binsz_wcs = u.Quantity(binsz_wcs)

        if wcs is None and region is not None:
            if isinstance(region, CompoundSkyRegion):
                self._center = compound_region_center(region)
            else:
                self._center = region.center

            wcs = WcsGeom.create(
                binsz=binsz_wcs,
                skydir=self._center,
                proj=self.projection,
                frame=self._center.frame.name,
            ).wcs
            self._wcs = wcs
            # TODO : can we get the width before defining the wcs ?
            wcs = WcsGeom.create(
                binsz=binsz_wcs,
                width=tuple(self.width),
                skydir=self._center,
                proj=self.projection,
                frame=self._center.frame.name,
            ).wcs

        self._wcs = wcs
        self.ndim = len(self.data_shape)

    @property
    def frame(self):
        """Coordinate system, either Galactic ("galactic") or Equatorial ("icrs")."""
        if self.region is None:
            return "icrs"
        try:
            return self.region.center.frame.name
        except AttributeError:
            return wcs_to_celestial_frame(self.wcs).name

    @property
    def binsz_wcs(self):
        """Angular bin size of the underlying `~WcsGeom`.

        Returns
        -------
        binsz_wcs: `~astropy.coordinates.Angle`
            Angular bin size.
        """
        return Angle(proj_plane_pixel_scales(self.wcs), unit="deg")

    @lazyproperty
    def _rectangle_bbox(self):
        if self.region is None:
            raise ValueError("Region definition required.")

        regions = compound_region_to_regions(self.region)
        regions_pix = [_.to_pixel(self.wcs) for _ in regions]

        bbox = regions_pix[0].bounding_box

        for region_pix in regions_pix[1:]:
            bbox = bbox.union(region_pix.bounding_box)

        try:
            rectangle_pix = bbox.to_region()
        except ValueError:
            rectangle_pix = RectanglePixelRegion(
                center=PixCoord(*bbox.center[::-1]), width=1, height=1
            )
        return rectangle_pix.to_sky(self.wcs)

    @property
    def width(self):
        """Width of bounding box of the region.

        Returns
        -------
        width : `~astropy.units.Quantity`
            Dimensions of the region in both spatial dimensions.
            Units: ``deg``
        """
        rectangle = self._rectangle_bbox
        return u.Quantity([rectangle.width.to("deg"), rectangle.height.to("deg")])

    @property
    def region(self):
        """The spatial component of the region geometry as a `~regions.SkyRegion`."""
        return self._region

    @property
    def is_all_point_sky_regions(self):
        """Whether regions are all point regions."""
        regions = compound_region_to_regions(self.region)
        return np.all([isinstance(_, PointSkyRegion) for _ in regions])

    @property
    def axes(self):
        """List of non-spatial axes."""
        return self._axes

    @property
    def axes_names(self):
        """All axes names."""
        return ["lon", "lat"] + self.axes.names

    @property
    def wcs(self):
        """WCS projection object."""
        return self._wcs

    @property
    def center_coord(self):
        """Center coordinate of the region as a `astropy.coordinates.SkyCoord`."""
        return self.pix_to_coord(self.center_pix)

    @property
    def center_pix(self):
        """Pixel values corresponding to the center of the region."""
        return tuple((np.array(self.data_shape) - 1.0) / 2)[::-1]

    @lazyproperty
    def center_skydir(self):
        """Sky coordinate of the center of the region."""
        if self.region is None:
            return SkyCoord(np.nan * u.deg, np.nan * u.deg)

        return self._rectangle_bbox.center

    @property
    def npix(self):
        """Number of spatial pixels."""
        return ([1], [1])

    def contains(self, coords):
        """Check if a given map coordinate is contained in the region.

        Requires the `.region` attribute to be set.
        For `PointSkyRegion` the method always returns True.

        Parameters
        ----------
        coords : tuple, dict, `MapCoord` or `~astropy.coordinates.SkyCoord`
            Object containing coordinate arrays we wish to check for inclusion
            in the region.

        Returns
        -------
        mask : `~numpy.ndarray`
            Boolean array with the same shape as the input that indicates
            which coordinates are inside the region.
        """
        if self.region is None:
            raise ValueError("Region definition required.")

        coords = MapCoord.create(coords, frame=self.frame, axis_names=self.axes.names)

        if self.is_all_point_sky_regions:
            return np.ones(coords.skycoord.shape, dtype=bool)

        return self.region.contains(coords.skycoord, self.wcs)

    def contains_wcs_pix(self, pix):
        """Check if a given WCS pixel coordinate is contained in the region.

        For `PointSkyRegion` the method always returns True.

        Parameters
        ----------
        pix : tuple
            Tuple of pixel coordinates.

        Returns
        -------
        containment : `~numpy.ndarray`
            Boolean array.
        """
        if self.is_all_point_sky_regions:
            return np.ones(pix[0].shape, dtype=bool)

        region_pix = self.region.to_pixel(self.wcs)
        return region_pix.contains(PixCoord(pix[0], pix[1]))

    def separation(self, position):
        """Angular distance between the center of the region and the given position.

        Parameters
        ----------
        position : `astropy.coordinates.SkyCoord`
            Sky coordinate we want the angular distance to.

        Returns
        -------
        sep : `~astropy.coordinates.Angle`
            The on-sky separation between the given coordinate and the region center.
        """
        return self.center_skydir.separation(position)

    @property
    def data_shape(self):
        """Shape of the `~numpy.ndarray` matching this geometry."""
        return self._shape[::-1]

    @property
    def data_shape_axes(self):
        """Shape of data of the non-spatial axes and unit spatial axes."""
        return self.axes.shape[::-1] + (1, 1)

    @property
    def _shape(self):
        """Number of bins in each dimension.

        The spatial dimension is always (1, 1), as a
        `RegionGeom` is not pixelized further.
        """
        return tuple((1, 1) + self.axes.shape)

    def get_coord(self, mode="center", frame=None, sparse=False, axis_name=None):
        """Get map coordinates from the geometry.

        Parameters
        ----------
        mode : {'center', 'edges'}, optional
            Get center or edge coordinates for the non-spatial axes.
            Default is "center".
        frame : str or `~astropy.coordinates.Frame`, optional
            Coordinate frame. Default is None.
        sparse : bool, optional
            Compute sparse coordinates. Default is False.
        axis_name : str, optional
            If mode = "edges", the edges will be returned for this axis only.
            Default is None.

        Returns
        -------
        coord : `~MapCoord`
            Map coordinate object.
        """
        if mode == "edges" and axis_name is None:
            raise ValueError("Mode 'edges' requires axis name")

        coords = self.axes.get_coord(mode=mode, axis_name=axis_name)
        coords["skycoord"] = self.center_skydir.reshape((1, 1))

        if frame is None:
            frame = self.frame

        coords = MapCoord.create(coords, frame=self.frame).to_frame(frame)

        if not sparse:
            coords = coords.broadcasted

        return coords

    def _pad_spatial(self, pad_width):
        raise NotImplementedError("Spatial padding of `RegionGeom` not supported")

    def crop(self):
        raise NotImplementedError("Cropping of `RegionGeom` not supported")

    def solid_angle(self):
        """Get solid angle of the region.

        Returns
        -------
        angle : `~astropy.units.Quantity`
            Solid angle of the region in steradians.
            Units: ``sr``
        """
        if self.region is None:
            raise ValueError("Region definition required.")

        # compound regions do not implement area()
        # so we use the mask representation and estimate the area
        # from the pixels in the mask using oversampling
        if isinstance(self.region, CompoundSkyRegion):
            # oversample by a factor of ten
            oversampling = 10.0
            wcs = self.to_binsz_wcs(self.binsz_wcs / oversampling).wcs
            pixel_region = self.region.to_pixel(wcs)
            mask = pixel_region.to_mask()
            area = np.count_nonzero(mask) / oversampling**2
        else:
            # all other types of regions should implement area
            area = self.region.to_pixel(self.wcs).area

        solid_angle = area * proj_plane_pixel_area(self.wcs) * u.deg**2
        return solid_angle.to("sr")

    def bin_volume(self):
        """If the `RegionGeom` has a non-spatial axis, it returns the volume of the region.
        If not, it returns the solid angle size.

        Returns
        -------
        volume : `~astropy.units.Quantity`
            Volume of the region.
        """
        bin_volume = self.solid_angle() * np.ones(self.data_shape)

        for idx, ax in enumerate(self.axes):
            shape = self.ndim * [1]
            shape[-(idx + 3)] = -1
            bin_volume = bin_volume * ax.bin_width.reshape(tuple(shape))

        return bin_volume

    def to_wcs_geom(self, width_min=None):
        """Get the minimal equivalent geometry which contains the region.

        Parameters
        ----------
        width_min : `~astropy.quantity.Quantity`, optional
            Minimum width for the resulting geometry. Can be a single number or two,
            for different minimum widths in each spatial dimension.
            Default is None.

        Returns
        -------
        wcs_geom : `~WcsGeom`
            A WCS geometry object.
        """
        if width_min is not None:
            width = np.max(
                [self.width.to_value("deg"), _check_width(width_min)], axis=0
            )
        else:
            width = self.width
        wcs_geom_region = WcsGeom(wcs=self.wcs)
        wcs_geom = wcs_geom_region.cutout(position=self.center_skydir, width=width)
        wcs_geom = wcs_geom.to_cube(self.axes)
        return wcs_geom

    def to_binsz_wcs(self, binsz):
        """Change the bin size of the underlying WCS geometry.

        Parameters
        ----------
        binsz : float, str or `~astropy.quantity.Quantity`
            Bin size.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the same axes and region as the input,
            but different WCS pixelization.
        """
        new_geom = RegionGeom(self.region, axes=self.axes, binsz_wcs=binsz)
        return new_geom

    @cachemethod
    def get_wcs_coord_and_weights(self, factor=10):
        """Get the array of spatial coordinates and corresponding weights.

        The coordinates are the center of a pixel that intersects the region and
        the weights that represent which fraction of the pixel is contained
        in the region.

        Parameters
        ----------
        factor : int, optional
            Oversampling factor to compute the weights.
            Default is 10.

        Returns
        -------
        region_coord : `~MapCoord`
            MapCoord object with the coordinates inside
            the region.
        weights : `~numpy.ndarray`
            Weights representing the fraction of each pixel
            contained in the region.
        """
        wcs_geom = self.to_wcs_geom()

        weights = wcs_geom.to_image().region_weights(
            regions=[self.region], oversampling_factor=factor
        )

        mask = weights.data > 0
        weights = weights.data[mask]

        # Get coordinates
        coords = wcs_geom.get_coord(sparse=True).apply_mask(mask)
        return coords, weights

    def to_binsz(self, binsz):
        """Return self."""
        return self

    def to_cube(self, axes):
        """Append non-spatial axes to create a higher-dimensional geometry.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the added axes.
        """
        axes = copy.deepcopy(self.axes) + axes
        return self._init_copy(region=self.region, wcs=self.wcs, axes=axes)

    def to_image(self):
        """Remove non-spatial axes to create a 2D region.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry without any non-spatial axes.
        """
        return self._init_copy(region=self.region, wcs=self.wcs, axes=None)

    def upsample(self, factor, axis_name=None):
        """Upsample a non-spatial dimension of the region by a given factor.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the upsampled axis.
        """
        axes = self.axes.upsample(factor=factor, axis_name=axis_name)
        return self._init_copy(region=self.region, wcs=self.wcs, axes=axes)

    def downsample(self, factor, axis_name):
        """Downsample a non-spatial dimension of the region by a given factor.

        Returns
        -------
        region : `~RegionGeom`
            Region geometry with the downsampled axis.
        """
        axes = self.axes.downsample(factor=factor, axis_name=axis_name)
        return self._init_copy(region=self.region, wcs=self.wcs, axes=axes)

    def pix_to_coord(self, pix):
        lon = np.where(
            (-0.5 < pix[0]) & (pix[0] < 0.5),
            self.center_skydir.data.lon,
            np.nan * u.deg,
        )
        lat = np.where(
            (-0.5 < pix[1]) & (pix[1] < 0.5),
            self.center_skydir.data.lat,
            np.nan * u.deg,
        )
        coords = (lon, lat)

        for p, ax in zip(pix[self._slice_non_spatial_axes], self.axes):
            coords += (ax.pix_to_coord(p),)

        return coords

    def pix_to_idx(self, pix, clip=False):
        idxs = list(pix_tuple_to_idx(pix))

        for i, idx in enumerate(idxs[self._slice_non_spatial_axes]):
            if clip:
                np.clip(idx, 0, self.axes[i].nbin - 1, out=idx)
            else:
                np.putmask(idx, (idx < 0) | (idx >= self.axes[i].nbin), -1)

        return tuple(idxs)

    def coord_to_pix(self, coords):
        # inherited docstring
        if isinstance(coords, tuple) and len(coords) == len(self.axes):
            skydir = self.center_skydir.transform_to(self.frame)
            coords = (skydir.data.lon, skydir.data.lat) + coords
        elif isinstance(coords, dict):
            valid_keys = ["lon", "lat", "skycoord"]
            if not any([_ in coords for _ in valid_keys]):
                coords.setdefault("skycoord", self.center_skydir)

        coords = MapCoord.create(coords, frame=self.frame, axis_names=self.axes.names)

        if self.region is None:
            pix = (0, 0)
        else:
            in_region = self.contains(coords.skycoord)

            x = np.zeros(coords.skycoord.shape)
            x[~in_region] = np.nan

            y = np.zeros(coords.skycoord.shape)
            y[~in_region] = np.nan

            pix = (x, y)

        pix += self.axes.coord_to_pix(coords)
        return pix

    def get_idx(self):
        idxs = [np.arange(n, dtype=float) for n in self.data_shape[::-1]]
        return np.meshgrid(*idxs[::-1], indexing="ij")[::-1]

    def _make_bands_cols(self):
        return []

    @classmethod
    def create(cls, region, **kwargs):
        """Create region geometry.

        The input region can be passed in the form of a ds9 string and will be parsed
        internally by `~regions.Regions.parse`. See:

        * https://astropy-regions.readthedocs.io/en/stable/region_io.html
        * http://ds9.si.edu/doc/ref/region.html

        Parameters
        ----------
        region : str or `~regions.SkyRegion`
            Region definition.
        **kwargs : dict
            Keyword arguments passed to `RegionGeom.__init__`.

        Returns
        -------
        geom : `RegionGeom`
            Region geometry.
        """
        return cls.from_regions(regions=region, **kwargs)

    def __str__(self):
        axes = ["lon", "lat"] + [_.name for _ in self.axes]
        try:
            frame = self.center_skydir.frame.name
            lon = self.center_skydir.data.lon.deg
            lat = self.center_skydir.data.lat.deg
        except AttributeError:
            frame, lon, lat = "", np.nan, np.nan

        return (
            f"{self.__class__.__name__}\n\n"
            f"\tregion     : {self.region.__class__.__name__}\n"
            f"\taxes       : {axes}\n"
            f"\tshape      : {self.data_shape[::-1]}\n"
            f"\tndim       : {self.ndim}\n"
            f"\tframe      : {frame}\n"
            f"\tcenter     : {lon:.1f} deg, {lat:.1f} deg\n"
        )

    def is_allclose(self, other, rtol_axes=1e-6, atol_axes=1e-6):
        """Compare two data IRFs for equivalency.

        Parameters
        ----------
        other : `RegionGeom`
            Region geometry to compare against.
        rtol_axes : float, optional
            Relative tolerance for the axes comparison.
            Default is 1e-6.
        atol_axes : float, optional
            Relative tolerance for the axes comparison.
            Default is 1e-6.

        Returns
        -------
        is_allclose : bool
            Whether the geometry is all close.
        """
        if not isinstance(other, self.__class__):
            return TypeError(f"Cannot compare {type(self)} and {type(other)}")

        if self.data_shape != other.data_shape:
            log.debug("RegionGeom data shape is not equal")
            return False

        axes_eq = self.axes.is_allclose(other.axes, rtol=rtol_axes, atol=atol_axes)
        if not axes_eq:
            log.debug("RegionGeom axes are not equal")
        # TODO: compare regions based on masks...
        regions_eq = True
        return axes_eq and regions_eq

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.is_allclose(other=other)

    def _to_region_table(self):
        """Export region to a FITS region table."""
        if self.region is None:
            raise ValueError("Region definition required.")

        region_list = compound_region_to_regions(self.region)

        pixel_region_list = []

        for reg in region_list:
            pixel_region_list.append(reg.to_pixel(self.wcs))

        table = Regions(pixel_region_list).serialize(format="fits")

        header = WcsGeom(wcs=self.wcs).to_header()
        table.meta.update(header)
        return table

    def to_hdulist(self, format="ogip", hdu_bands=None, hdu_region=None):
        """Convert geometry to HDU list.

        Parameters
        ----------
        format : {"ogip", "gadf", "ogip-sherpa"}
            HDU format. Default is "ogip".
        hdu_bands : str, optional
            Name or index of the HDU with the BANDS table.
            Default is None.
        hdu_region : str, optional
            Name or index of the HDU with the region table.
            Not used for the "gadf" format.
            Default is None.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
            HDU list.

        """
        if hdu_bands is None:
            hdu_bands = "HDU_BANDS"
        if hdu_region is None:
            hdu_region = "HDU_REGION"
        if format != "gadf":
            hdu_region = "REGION"

        hdulist = fits.HDUList()

        hdulist.append(self.axes.to_table_hdu(hdu_bands=hdu_bands, format=format))

        # region HDU
        if self.region:
            region_table = self._to_region_table()

            region_hdu = fits.BinTableHDU(region_table, name=hdu_region)
            hdulist.append(region_hdu)

        return hdulist

    @classmethod
    def from_regions(cls, regions, **kwargs):
        """Create region geometry from list of regions.

        The regions are combined with union to a compound region.

        Parameters
        ----------
        regions : list of `~regions.SkyRegion` or str
            Regions.
        **kwargs: dict
            Keyword arguments forwarded to `RegionGeom`.

        Returns
        -------
        geom : `RegionGeom`
            Region map geometry.
        """
        regions = _parse_regions(regions)

        if regions:
            regions = regions_to_compound_region(regions)

        return cls(region=regions, **kwargs)

    @classmethod
    def from_hdulist(cls, hdulist, format="ogip", hdu=None):
        """Read region table and convert it to region list.

        Parameters
        ----------
        hdulist : `~astropy.io.fits.HDUList`
            HDU list.
        format : {"ogip", "ogip-arf", "gadf"}
            HDU format. Default is "ogip".
        hdu : str, optional
            Name of the HDU. Default is None.

        Returns
        -------
        geom : `RegionGeom`
            Region map geometry.

        """
        region_hdu = "REGION"

        if format == "gadf" and hdu:
            region_hdu = f"{hdu}_{region_hdu}"

        if region_hdu in hdulist:
            try:
                region_table = QTable.read(hdulist[region_hdu])
                regions_pix = Regions.parse(data=region_table, format="fits")
            except TypeError:
                region_table = Table.read(hdulist[region_hdu])
                regions_pix = Regions.parse(data=region_table, format="fits")

            wcs = WcsGeom.from_header(region_table.meta).wcs
            regions = []

            for region_pix in regions_pix:
                # TODO: remove workaround once regions issue with fits serialization is sorted out
                # see https://github.com/astropy/regions/issues/400
                # requires update regions to >=v0.6
                region_pix.meta["include"] = True
                regions.append(region_pix.to_sky(wcs))

            region = regions_to_compound_region(regions)
        else:
            region, wcs = None, None

        if format == "ogip":
            hdu_bands = "EBOUNDS"
        elif format == "ogip-arf":
            hdu_bands = "SPECRESP"
        elif format == "gadf":
            hdu_bands = hdu + "_BANDS"
        else:
            raise ValueError(f"Unknown format {format}")

        axes = MapAxes.from_table_hdu(hdulist[hdu_bands], format=format)
        return cls(region=region, wcs=wcs, axes=axes)

    def union(self, other):
        """Stack a `RegionGeom` by making the union."""
        if not self == other:
            raise ValueError("Can only make union if extra axes are equivalent.")
        if other.region:
            if self.region:
                self._region = self.region.union(other.region)
            else:
                self._region = other.region

    def plot_region(self, ax=None, kwargs_point=None, path_effect=None, **kwargs):
        """Plot region in the sky.

        Parameters
        ----------
        ax : `~astropy.visualization.WCSAxes`, optional
            Axes to plot on. If no axes are given,
            the region is shown using the minimal
            equivalent WCS geometry.
            Default is None.
        kwargs_point : dict, optional
            Keyword arguments passed to `~matplotlib.lines.Line2D` for plotting
            of point sources. Default is None.
        path_effect : `~matplotlib.patheffects.PathEffect`, optional
            Path effect applied to artists and lines.
            Default is None.
        **kwargs : dict
            Keyword arguments forwarded to `~regions.PixelRegion.as_artist`.

        Returns
        -------
        ax : `~astropy.visualization.WCSAxes`
            Axes to plot on.
        """
        if self.region:
            kwargs_point = kwargs_point or {}

            if ax is None:
                ax = plt.gca()

                if not isinstance(ax, WCSAxes):
                    ax.remove()
                    wcs_geom = self.to_wcs_geom()
                    m = Map.from_geom(geom=wcs_geom.to_image())
                    ax = m.plot(add_cbar=False, vmin=-1, vmax=0)

            kwargs.setdefault("facecolor", "None")
            kwargs.setdefault("edgecolor", "tab:blue")
            kwargs_point.setdefault("marker", "*")

            for key, value in kwargs.items():
                key_point = ARTIST_TO_LINE_PROPERTIES.get(key, None)
                if key_point and key_point not in kwargs_point:
                    kwargs_point[key_point] = value

            if "color" in kwargs:
                kwargs_point.setdefault("color", kwargs["color"])
            if "color" in kwargs_point:
                kwargs_point["markeredgecolor"] = kwargs_point["color"]

            for region in compound_region_to_regions(self.region):
                region_pix = region.to_pixel(wcs=ax.wcs)

                if isinstance(region, PointSkyRegion):
                    artist = region_pix.as_artist(**kwargs_point)
                else:
                    artist = region_pix.as_artist(**kwargs)

                if path_effect:
                    artist.add_path_effect(path_effect)

                ax.add_artist(artist)

            return ax
        else:
            logging.info("Region definition required.")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Table helper utilities."""

import numpy as np
from astropy.table import Table
from astropy.units import Quantity
from .units import standardise_unit

__all__ = [
    "hstack_columns",
    "table_row_to_dict",
    "table_standardise_units_copy",
    "table_standardise_units_inplace",
]


def hstack_columns(table, table_other):
    """Stack the column data horizontally.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Input table.
    table_other : `~astropy.table.Table`
        Other input table.

    Returns
    -------
    stacked : `~astropy.table.Table`
        Stacked table.
    """
    stacked = Table()

    for column in table.colnames:
        data = np.hstack([table[column].data[0], table_other[column].data[0]])
        stacked[column] = data[np.newaxis, :]
    return stacked


def table_standardise_units_copy(table):
    """Standardise units for all columns in a table in a copy.

    Calls `~gammapy.utils.units.standardise_unit`.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Input table (won't be modified).

    Returns
    -------
    table : `~astropy.table.Table`
        Copy of the input table with standardised column units.
    """
    # Note: we could add an `inplace` option (or variant of this function)
    # See https://github.com/astropy/astropy/issues/6098
    table = Table(table)
    return table_standardise_units_inplace(table)


def table_standardise_units_inplace(table):
    """Standardise units for all columns in a table in place."""
    for column in table.columns.values():
        if column.unit:
            column.unit = standardise_unit(column.unit)

    return table


def table_row_to_dict(row, make_quantity=True):
    """Make one source data dictionary.

    Parameters
    ----------
    row : `~astropy.table.Row`
        Row.
    make_quantity : bool, optional
        Make quantity values for columns with units.
        Default is True.

    Returns
    -------
    data : dict
        Row data.
    """
    data = {}
    for name, col in row.columns.items():
        val = row[name]

        is_str = np.issubdtype(col.dtype, np.str_) | np.issubdtype(col.dtype, np.bytes_)
        if isinstance(val, np.ma.core.MaskedConstant) and is_str:
            val = str(val)

        if make_quantity and col.unit:
            val = Quantity(val, unit=col.unit)
        data[name] = val
    return data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from astropy.io import fits


def find_bands_hdu(hdu_list, hdu):
    """Discover the extension name of the BANDS HDU.

    Parameters
    ----------
    hdu_list : `~astropy.io.fits.HDUList`
        The FITS HDU list.
    hdu : `~astropy.io.fits.BinTableHDU` or `~astropy.io.fits.ImageHDU`
        The HDU to check.

    Returns
    -------
    hduname : str
        Extension name of the BANDS HDU. Returns None if no BANDS HDU was found.
    """
    if "BANDSHDU" in hdu.header:
        return hdu.header["BANDSHDU"]

    has_cube_data = False

    if (
        isinstance(hdu, (fits.ImageHDU, fits.PrimaryHDU))
        and hdu.header.get("NAXIS", None) == 3
    ):
        has_cube_data = True
    elif isinstance(hdu, fits.BinTableHDU):
        if (
            hdu.header.get("INDXSCHM", "") in ["EXPLICIT", "IMPLICIT", ""]
            and len(hdu.columns) > 1
        ):
            has_cube_data = True

    if has_cube_data:
        if "EBOUNDS" in hdu_list:
            return "EBOUNDS"
        elif "ENERGIES" in hdu_list:
            return "ENERGIES"

    return None


def find_hdu(hdulist):
    """Find the first non-empty HDU."""
    for hdu in hdulist:
        if hdu.data is not None:
            return hdu

    raise AttributeError("No Image or BinTable HDU found.")


def find_bintable_hdu(hdulist):
    for hdu in hdulist:
        if hdu.data is not None and isinstance(hdu, fits.BinTableHDU):
            return hdu

    raise AttributeError("No BinTable HDU found.")



"""
HAWC with Gammapy
=================

Explore HAWC event lists and instrument response functions (IRFs), then perform the
data reduction steps.

Introduction
------------

`HAWC <https://www.hawc-observatory.org/>`__ is an array of
water Cherenkov detectors located in Mexico that detects gamma-rays
in the range between hundreds of GeV and hundreds of TeV.
Gammapy recently added support of HAWC high level data analysis,
after export to the current `open data level 3
format <https://gamma-astro-data-formats.readthedocs.io/>`__.

The HAWC data is largely private. However, in 2022, a small
sub-set of HAWC Pass4 event lists from the Crab nebula region
was publicly `released <https://data.hawc-observatory.org/datasets/crab_events_pass4/index.php>`__.
This dataset is 1 GB in size, so only a subset will be used here as an example.

Tutorial overview
-----------------

This notebook is a quick introduction to HAWC data analysis with Gammapy.
It briefly describes the HAWC data and how to access it, and then uses a
subset of the data to produce a `~gammapy.datasets.MapDataset`, to show how the
data reduction is performed.

The HAWC data release contains events where the energy is estimated using
two different algorithms, referred to as "NN" and "GP" (see this
`paper <https://iopscience.iop.org/article/10.3847/1538-4357/ab2f7d>`__
for a detailed description). These two event classes are not independent, meaning that
events are repeated between the NN and GP datasets. Therefore, these data should never
be analysed jointly, and one of the two estimators needs to be chosen before
proceeding.

Once the data has been reduced to a `~gammapy.datasets.MapDataset`, there are no differences
in the way that HAWC data is handled with respect to data from any other
observatory, such as H.E.S.S. or CTAO.


HAWC data access and reduction
------------------------------

This is how to access data and IRFs from the HAWC Crab event data release.

"""

import astropy.units as u
from astropy.coordinates import SkyCoord
import matplotlib.pyplot as plt
from gammapy.data import DataStore, HDUIndexTable, ObservationTable
from gammapy.datasets import MapDataset
from gammapy.estimators import ExcessMapEstimator
from gammapy.makers import MapDatasetMaker, SafeMaskMaker
from gammapy.maps import Map, MapAxis, WcsGeom


######################################################################
# Chose which estimator we will use

energy_estimator = "NN"


######################################################################
# A useful way to organize the relevant files are with index tables. The
# observation index table contains information on each particular observation,
# such as the run ID. The HDU index table has a row per
# relevant file (i.e., events, effective area, psf…) and contains the path
# to said file.
# The HAWC data is divided into different event types, classified using
# the fraction of the array that was triggered by an event, a quantity
# usually referred to as "fHit". These event types are fully independent,
# meaning that an event will have a unique event type identifier, which
# is usually a number indicating which fHit bin the event corresponds to.
# For this reason, a single HAWC observation has several HDU index tables
# associated to it, one per event type. In each table, there will be
# paths to a distinct event list file and IRFs.
# In the HAWC event data release, all the HDU tables are saved into
# the same FITS file, and can be accesses through the choice of the hdu index.


######################################################################
# Load the tables
# ~~~~~~~~~~~~~~~

data_path = "$GAMMAPY_DATA/hawc/crab_events_pass4/"
hdu_filename = f"hdu-index-table-{energy_estimator}-Crab.fits.gz"
obs_filename = f"obs-index-table-{energy_estimator}-Crab.fits.gz"

######################################################################
# There is only one observation table
obs_table = ObservationTable.read(data_path + obs_filename)

######################################################################
# The remainder of this tutorial utilises just one fHit value, however,
# for a regular analysis you should combine all fHit bins. Here,
# we utilise fHit bin number 6. We start by reading the HDU index table
# of this fHit bin

fHit = 6
hdu_table = HDUIndexTable.read(data_path + hdu_filename, hdu=fHit)


######################################################################
# From the tables, we can create a `~gammapy.data.DataStore`.

data_store = DataStore(hdu_table=hdu_table, obs_table=obs_table)

data_store.info()


######################################################################
# There is only one observation

obs = data_store.get_observations()[0]

######################################################################
# Peek events from this observation

obs.events.peek()
plt.show()


######################################################################
# Peek the energy dispersion:

obs.edisp.peek()
plt.show()

######################################################################
# Peek the psf:
obs.psf.peek()
plt.show()

######################################################################
# Peek the background for one transit:
plt.figure()
obs.bkg.reduce_over_axes().plot(add_cbar=True)
plt.show()

######################################################################
# Peek the effective exposure for one transit:
plt.figure()
obs.aeff.reduce_over_axes().plot(add_cbar=True)
plt.show()


######################################################################
# Data reduction into a `~gammapy.datasets.MapDataset`
# ----------------------------------------------------
#
# We will now produce a `~gammapy.datasets.MapDataset` using the data from one of the
# fHit bins. In order to use all bins, one just needs to repeat this
# process inside a for loop that modifies the variable fHit.


######################################################################
# First we define the geometry and axes, starting with the energy reco axis:

energy_axis = MapAxis.from_edges(
    [1.00, 1.78, 3.16, 5.62, 10.0, 17.8, 31.6, 56.2, 100, 177, 316] * u.TeV,
    name="energy",
    interp="log",
)

######################################################################
# Note: this axis is the one used to create the background model map. If
# different edges are used, the `~gammapy.makers.MapDatasetMaker` will interpolate between
# them, which might lead to unexpected behaviour.

######################################################################
# Define the energy true axis:

energy_axis_true = MapAxis.from_energy_bounds(
    1e-3, 1e4, nbin=140, unit="TeV", name="energy_true"
)

######################################################################
# Finally, create a geometry around the Crab location:

geom = WcsGeom.create(
    skydir=SkyCoord(ra=83.63, dec=22.01, unit="deg", frame="icrs"),
    width=6 * u.deg,
    axes=[energy_axis],
    binsz=0.05,
)


######################################################################
# Define the makers we will use:

maker = MapDatasetMaker(selection=["counts", "background", "exposure", "edisp", "psf"])
safe_mask_maker = SafeMaskMaker(methods=["aeff-max"], aeff_percent=10)


######################################################################
# Create an empty `~gammapy.datasets.MapDataset`.
# The keyword ``reco_psf=True`` is needed because the HAWC PSF is
# derived in reconstructed energy.

dataset_empty = MapDataset.create(
    geom, energy_axis_true=energy_axis_true, name=f"fHit {fHit}", reco_psf=True
)
dataset = maker.run(dataset_empty, obs)

######################################################################
# The livetime information is used by the `~gammapy.makers.SafeMaskMaker` to retrieve the
# effective area from the exposure. The HAWC effective area is computed
# for one source transit above 45º zenith, which is around 6h.
# Since the effective area condition used here is relative to
# the maximum, this value does not actually impact the result.

dataset.exposure.meta["livetime"] = "6 h"
dataset = safe_mask_maker.run(dataset)


######################################################################
# Now we have a dataset that has background and exposure quantities for
# one single transit, but our dataset might comprise more. The number
# of transits can be derived using the good time intervals (GTI) stored
# with the event list. For convenience, the HAWC data release already
# included this quantity as a map.

transit_map = Map.read(data_path + "irfs/TransitsMap_Crab.fits.gz")
transit_number = transit_map.get_by_coord(geom.center_skydir)

######################################################################
# Correct the background and exposure quantities:
dataset.background.data *= transit_number
dataset.exposure.data *= transit_number


######################################################################
# Check the dataset we produced
# -----------------------------
#
# We will now check the contents of the dataset.
# We can use the ``.peek()`` method to quickly get a glimpse of the contents
dataset.peek()
plt.show()


######################################################################
# Create significance maps to check that the Crab is visible:

excess_estimator = ExcessMapEstimator(
    correlation_radius="0.2 deg", selection_optional=[], energy_edges=energy_axis.edges
)
excess = excess_estimator.run(dataset)

(dataset.mask * excess["sqrt_ts"]).plot_grid(
    add_cbar=True, cmap="coolwarm", vmin=-5, vmax=5
)
plt.show()

######################################################################
# Combining all energies

excess_estimator_integrated = ExcessMapEstimator(
    correlation_radius="0.2 deg", selection_optional=[]
)
excess_integrated = excess_estimator_integrated.run(dataset)

excess_integrated["sqrt_ts"].plot(add_cbar=True)
plt.show()


######################################################################
# Exercises
# ---------
#
# -  Repeat the process for a different fHit bin.
# -  Repeat the process for all the fHit bins provided in the data
#    release and fit a model to the result.
#


######################################################################
# Next steps
# ----------
#
# Now you know how to access and work with HAWC data. All other
# tutorials and documentation concerning 3D analysis techniques and
# the `~gammapy.datasets.MapDataset` object can be used from this step on.
#
